# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
 Helper functions and nipype interface for
 reading slice timing correction parameters specially from Siemens acquisitions
"""
import os.path as op

import numpy as np
import nibabel as nib
from nipype.interfaces.base import (BaseInterface,
                                    TraitedSpec,
                                    InputMultiPath,
                                    BaseInterfaceInputSpec,
                                    traits,)

from .._utils        import check_equal, grep
from ..utils         import get_trait_value
from ..preproc.dicom import split_dcm_ahdr, dcm_ascii_hdr


def slicing_mode(dcm_file):
    """ Return the slicing mode of the fMRI acquisition file given
    one of its DICOM files. Avoid giving the first DICOM file.

    Parameters
    ----------
    dcm_file: str
       Path to the DICOM file

    Returns
    -------
    mode: str
        Choices: ('ascending', 'descending', 'interleaved')

    References
    ----------
    https://wiki.cimec.unitn.it/tiki-index.php?page=MRIBOLDfMRI
    http://www.mccauslandcenter.sc.edu/CRNL/tools/stc

    https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=ind03&L=SPM&D=0&P=5914721
    The timing can be found out most easily by looking into some shadow
    information: In any image from the syngo systems the measurement-protocol is
    present (in ascii). You may find it by looking for "### ASCCONV BEGIN ###"
    and "### ASCCONV END ###". Here you may also find the slice positions of the
    single images.
    > Now find "sSliceArray.ucMode". This is
    > 0x1 for ascending
    > 0x2 for descending
    > 0x4 for interleaved
    > In the interleaved mode, the information given by Peter Erhard is correct;
    for the rest it's clear anyway.

    Example
    -------
    import os.path as op
    from glob import glob
    dcmdir   = '/home/alexandre/data/pet_stadhauders/hiswork/FTD/'
    restdirs = [glob(op.join(op.abspath(sd), '*ep2d*')) for sd in glob(op.join(dcmdir, '*'))]
    dcms    = [glob(op.join(rd[0], '*.dcm'))[10] for rd in restdirs if rd]
    modes   = [slicing_mode(dcm) for dcm in dcms]
    print(modes)
    """
    _, ascconv = split_dcm_ahdr(dcm_ascii_hdr(dcm_file))

    mode_code = grep(ascconv, 'sSliceArray.ucMode')[0].split('=')[1].strip()

    code_modes = {'0x1': 'ascending',
                  '0x2': 'descending',
                  '0x4': 'interleaved'}

    return code_modes[mode_code]


def _get_n_slices(in_file):
    img = nib.load(in_file)

    n_slices = 0
    try:
        n_slices = img.header.get_n_slices()
    except:
        n_slices = img.shape[2]
    finally:
        return n_slices


def _get_time_repetition(in_file):
    tr = nib.load(in_file).header.get_dim_info()[2]

    if tr is None:
        raise ValueError('Could not pick the TR value from {}. Please set time_repetition in the configuration.')

    if tr < 1.0 or tr > 5.0:
        raise ValueError('Aborting: strange Repeat Time (TR), got {}.'.format(tr))
    if  tr > 5.0:
        raise ValueError('Long TR often used with sparse imaging: if this is a sparse design please set the TR manually.')
    if  tr < 0.5:
        raise ValueError('Short TR may be due to DICOM-to-NIfTI conversion. Perhaps use dcm2nii.')

    return tr


def _get_time_acquisition(in_file, TR, n_slices):
    if n_slices <= 0: # not only to avoid division by zero
        raise ValueError('Null number of slices when calculating time '
                         'acquisition for {}, got {}. Please set num_slices '
                         'in the configuration.'.format(in_file, n_slices))

    return (TR/n_slices) * (n_slices-1)


def _get_ref_slice(in_file, slice_order):
    if slice_order is None:
        raise ValueError('Expected a list of integers as `slice_order`, got {}.'.format(slice_order))

    return slice_order[0]


def _get_slice_order(in_file, n_slices, slice_mode):

    def read_slice_mode_byte(in_file):
        try:
            with open(in_file) as f:
                f.seek(122)
                slice_mode = f.read(1)
        except:
            return -1
        else:
            return slice_mode

    def get_nii_slice_times(img):
        # try get the slice times
        try:
            times = img.header.get_slice_times()
        except Exception:
            pass
        else:
            return times

    def order_from_times(times):
        return np.argsort(times) + 1

    def calculate_slice_order(n_slices, slice_mode):
        """

        Parameters
        ----------
        n_slices: int

        slice_mode: int or str
            #  0: 'unknown' : ask for automatic detection of the slice order
            #  1: 'seq_inc' : sequential ascending kNIFTI_SLICE_SEQ_INC = 1; %1,2,3,4
            #  2: 'seq_dec' : sequential descending kNIFTI_SLICE_SEQ_DEC = 2; %4,3,2,1
            #  3: 'alt_inc' : Siemens: interleaved ascending with odd number of slices,
                                       interleaved for other vendors kNIFTI_SLICE_ALT_INC = 3; %1,3,2,4
            #  4: 'alt_dec' : descending interleaved kNIFTI_SLICE_ALT_DEC = 4; %4,2,3,1
            #  5: 'alt_inc2': Siemens interleaved ascending with even number of slices kNIFTI_SLICE_ALT_INC2 = 5; %2,4,1,3
            #  6: 'alt_dec2': Siemens interleaved descending with even number of slices kNIFTI_SLICE_ALT_DEC2 = 6; %3,1,4,2

        Returns
        -------
        slice_order: list of int
        """
        mode_int = { 0: 'unknown',
                     1: 'seq_inc',
                     2: 'seq_dec',
                     3: 'alt_inc',
                     4: 'alt_dec',
                     5: 'alt_inc2',
                     6: 'alt_dec2',}

        if isinstance(slice_mode, int):
            slice_mode = mode_int[slice_mode]

        choices = tuple(mode_int.values())
        if slice_mode not in choices:
            raise ValueError('Expected `slice_mode` to be in {}, got {}.'.format(choices,
                                                                                 slice_mode))

        is_siemens = False
        if slice_mode in ('alt_inc2', 'alt_dec2'):
            is_siemens = True

        if 'seq' in slice_mode: # sequential
            slice_order = list(range(n_slices))
        else: # interleaved
            if is_siemens and '2' in slice_mode: #siemens and even number of slices
                slice_order = list(range(1, n_slices, 2)) + list(range(0, n_slices, 2))
            else:
                slice_order = list(range(0, n_slices, 2)) + list(range(1, n_slices, 2))

        if 'dec' in slice_mode: # descending
            slice_order = [n_slices - 1 - i for i in slice_order]

        return slice_order

    # starts the function
    if slice_mode == 'unknown':
        # check if the slice times are in the NifTI header
        img = nib.load(in_file)
        times = get_nii_slice_times(img)
        if times is not None:
            return order_from_times(times)

    # read the slice mode code from the file
    if slice_mode == 'unknown':
        slice_mode = read_slice_mode_byte(in_file)
        if slice_mode <= 0:
            slice_mode = 'unknown'

    if slice_mode == 'unknown':
        raise AttributeError("Don't have enough information to calculate the "
                             "slice order from {}. Please set the slice_mode argument.".format(in_file))

    return calculate_slice_order(n_slices, slice_mode)


class STCParameters(object):
    """ Class to calculate the parameters needed for slice timing correction.
    Some options are automated for Siemens acquisitions.

    Auto detection of slice order, i.e., slice_order == [0] and slice_mode == 'unknown'
    only works for images from Siemens and converted with dcm2nii from Nov 2013 or later.

    Siemens have unusual interleaving
    - http://cbs.fas.harvard.edu/node/559#slice_order
    - https://wiki.cimec.unitn.it/tiki-index.php?page=MRIBOLDfMRI

    This class is based on the script by Chris Rorden's Neuropsychology Lab in:
    - http://www.mccauslandcenter.sc.edu/CRNL/tools/stc

    This is a callable class so you can instance this class in a nipype Function object.
    See `slice_timing_params` in this module after the declaration of this class.
    """
    def __init__(self):
        self.in_files         = None
        self.num_slices       = 0
        self.slice_order      = None
        self.time_repetition  = None
        self.time_acquisition = None
        self.ref_slice        = None
        self.slice_mode       = 'unknown'

    def fit(self, in_files,
                  num_slices       = 0,
                  ref_slice        = None,
                  slice_order      = None,
                  time_acquisition = None,
                  time_repetition  = None,
                  slice_mode       = 'unknown'):
        """

        Parameters
        ----------
        in_files: str
            Path to the input files

        num_slices: int
            Number of slices of `in_files`.

        ref_slice: int
            Index of the reference slice

        slice_order: list of ints
            List of integers with the order in which slices are acquired

        time_acquisition: int
            Time of volume acquisition. usually calculated as TR-(TR/num_slices)

        time_repetition: int
            The time repetition (TR) of the input dataset in seconds
            Default: 0
            If left to default will read the TR from the nifti image header.

        slice_mode: str
            Choices:
                'unknown': auto detect if images are from Siemens and converted with dcm2nii from Nov 2013 or later #kNIFTI_SLICE_UNKNOWN
                'seq_inc': sequential ascending kNIFTI_SLICE_SEQ_INC = 1; %1,2,3,4
                'seq_dec': sequential descending kNIFTI_SLICE_SEQ_DEC = 2; %4,3,2,1
                'alt_inc': Siemens: interleaved ascending with odd number of slices,
                                    interleaved for other vendors kNIFTI_SLICE_ALT_INC = 3; %1,3,2,4
                'alt_dec': descending interleaved kNIFTI_SLICE_ALT_DEC = 4; %4,2,3,1
                'alt_inc2': Siemens interleaved ascending with even number of slices kNIFTI_SLICE_ALT_INC2 = 5; %2,4,1,3
                'alt_dec2': Siemens interleaved descending with even number of slices kNIFTI_SLICE_ALT_DEC2 = 6; %3,1,4,2

            Default: 'unknown'
            If left to default will try to detect the TR from the nifti image header, if it doesn't work
            an AttributeError exception will be raise.

        Returns
        -------
        num_slices

        ref_slice

        slice_order

        time_acquisition

        time_repetition
        """
        self.in_files         = in_files
        self.num_slices       = num_slices
        self.slice_order      = slice_order
        self.time_repetition  = time_repetition
        self.time_acquisition = time_acquisition
        self.ref_slice        = ref_slice
        self.slice_mode       = slice_mode

        _ = self.set_num_slices()
        _ = self.set_slice_order()
        _ = self.set_time_repetition()
        _ = self.set_time_acquisition()
        _ = self.set_ref_slice()

        return (self.num_slices,
                self.ref_slice,
                self.slice_order,
                self.time_acquisition,
                self.time_repetition)

    def _check_in_files(self):
        if isinstance(self.in_files, str):
            in_files = [self.in_files]
        else:
            in_files = self.in_files

        for f in in_files:
            if not op.exists(f):
                raise IOError('Expected an existing file in `in_files`')

        return in_files

    def _check_all_equal(self, func, error_msg=None, **kwargs):
        in_files = self._check_in_files()

        values = [func(f,  **kwargs) for f in in_files]
        if not check_equal(values):
            if error_msg is None:
                error_msg = 'The values from {} are not the same, got {{}}.'.format(func.__name__)
            raise ValueError(error_msg.format(values))

        return values[0]

    def set_time_acquisition(self):
        if self.time_acquisition is not None:
            return self.time_acquisition

        n_slices = self.set_num_slices()
        tr = self.set_time_repetition()

        error_msg = 'The time acquisition calculated from all the `in_files` are not the same, got {}.'
        self.time_acquisition = self._check_all_equal(_get_time_acquisition,
                                                      error_msg,
                                                      TR=tr,
                                                      n_slices=n_slices)
        return self.time_acquisition

    def set_num_slices(self):
        if self.num_slices is not None:
            if self.num_slices > 0:
                return self.num_slices

        error_msg = 'The number of z slices for all `in_files` are not the same, got {}.'
        self.num_slices = self._check_all_equal(_get_n_slices, error_msg)
        return self.num_slices

    def set_time_repetition(self):
        if self.time_repetition is not None:
            return self.time_repetition

        error_msg = 'The TR calculated from all the `in_files` are not the same, got {}.'
        self.time_repetition = self._check_all_equal(_get_time_repetition, error_msg)
        return self.time_repetition

    def set_slice_order(self):
        if self.slice_order is not None:
            return self.slice_order

        n_slices = self.set_num_slices()

        error_msg = 'The slice order for all `in_files` are not the same, got {}.'
        self.slice_order = self._check_all_equal(_get_slice_order,
                                                 error_msg,
                                                 n_slices=n_slices,
                                                 slice_mode=self.slice_mode)

        return self.slice_order

    def set_ref_slice(self):
        """set slice order to the first slice http://www.alivelearn.net/?p=1037
           slice_order[0] is first acquired slice, set it as the refernece
        """
        if self.ref_slice is not None:
            return self.slice_order

        slice_order = self.set_slice_order()

        error_msg = 'The reference slice for all `in_files` are not the same, got {}.'
        self.ref_slice = self._check_all_equal(_get_ref_slice, error_msg,
                                               slice_order=slice_order)
        return self.ref_slice


class STCParametersInputSpec(BaseInterfaceInputSpec):
    in_files         = InputMultiPath(traits.File(desc="fMRI image file(s) from where to obtain the slice time "
                                                       "correction parameters.", exists=True, mandatory=True))
    num_slices       = traits.Int    (desc="Number of slices (depends on acquisition direction).")
    slice_order      = traits.ListInt(desc="List of integers with the order in which slices are acquired.")
    time_repetition  = traits.Float  (desc="The time repetition (TR) of the input dataset in seconds. "
                                           "If left to default will read the TR from the nifti image header.", default=0)
    time_acquisition = traits.Float  (desc="Time of volume acquisition. usually calculated as TR-(TR/num_slices).")
    ref_slice        = traits.Int    (desc="Index of the slice of reference.")
    slice_mode       = traits.Enum   ('unknown', 'seq_inc', 'seq_dec', 'alt_inc', 'alt_dec', 'alt_inc2', 'alt_dec2',
                                      desc="Slicing mode of the acquisition. \n"
                                      "Choices: \n"
                                      "    'unknown': auto detect if images are from Siemens and converted with dcm2nii"
                                      "               from Nov 2013 or later #kNIFTI_SLICE_UNKNOWN\n"
                                      "    'seq_inc': sequential ascending kNIFTI_SLICE_SEQ_INC = 1; %1,2,3,4\n"
                                      "    'seq_dec': sequential descending kNIFTI_SLICE_SEQ_DEC = 2; %4,3,2,1\n"
                                      "    'alt_inc': Siemens: interleaved ascending with odd number of slices,"
                                      "                        interleaved for other vendors kNIFTI_SLICE_ALT_INC = 3; %1,3,2,4\n"
                                      "    'alt_dec': descending interleaved kNIFTI_SLICE_ALT_DEC = 4; %4,2,3,1\n"
                                      "    'alt_inc2': Siemens interleaved ascending with even number of slices kNIFTI_SLICE_ALT_INC2 = 5; %2,4,1,3\n"
                                      "    'alt_dec2': Siemens interleaved descending with even number of slices kNIFTI_SLICE_ALT_DEC2 = 6; %3,1,4,2\n"
                                      "If left to default will try to detect the TR from the nifti image header, if it doesn't work"
                                      "an AttributeError exception will be raise.", default='unknown')


class STCParametersOutputSpec(TraitedSpec):
    in_files         = InputMultiPath(traits.File(desc="fMRI image file from where to obtain the slice time correction "
                                                       "parameters.", exists=True, mandatory=True))
    num_slices       = traits.Int    (desc="Number of slices (depends on acquisition direction).")
    slice_order      = traits.ListInt(desc="List of integers with the order in which slices are acquired.")
    time_repetition  = traits.Float  (desc="The time repetition (TR) of the input dataset in seconds. "
                                           "If left to default will read the TR from the nifti image header.", default=0)
    time_acquisition = traits.Float  (desc="Time of volume acquisition. usually calculated as TR-(TR/num_slices).")
    ref_slice        = traits.Int    (desc="Index of the slice of reference.")


class STCParametersInterface(BaseInterface):
    """
    Class to calculate the parameters needed for slice timing correction.
    Some options are automated for Siemens acquisitions.

    Check the STCParameters class docstring for more info.
    """
    input_spec = STCParametersInputSpec
    output_spec = STCParametersOutputSpec

    def _run_interface(self, runtime):

        num_slices       = get_trait_value(self.inputs, 'num_slices',       default=0)
        ref_slice        = get_trait_value(self.inputs, 'ref_slice',        default=None)
        slice_order      = get_trait_value(self.inputs, 'slice_order',      default=None)
        time_acquisition = get_trait_value(self.inputs, 'time_acquisition', default=None)
        time_repetition  = get_trait_value(self.inputs, 'time_repetition',  default=None)
        slice_mode       = get_trait_value(self.inputs, 'slice_mode',       default='unknown')

        self.stc_params = STCParameters()

        (self._num_slices,
         self._ref_slice,
         self._slice_order,
         self._time_acquisition,
         self._time_repetition) = self.stc_params.fit(in_files         = self.inputs.in_files,
                                                      num_slices       = num_slices,
                                                      ref_slice        = ref_slice,
                                                      slice_order      = slice_order,
                                                      time_acquisition = time_acquisition,
                                                      time_repetition  = time_repetition,
                                                      slice_mode       = slice_mode,
                                                      )
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['in_files'        ] = self.inputs.in_files
        outputs['num_slices'      ] = self._num_slices
        outputs['ref_slice'       ] = self._ref_slice
        outputs['slice_order'     ] = self._slice_order
        outputs['time_acquisition'] = self._time_acquisition
        outputs['time_repetition' ] = self._time_repetition
        return outputs



# -*- coding: utf-8 -*-
"""
Configuration manager for workflow definitions.
This works over Kaptan:https://github.com/emre/kaptan

The global configuration registry is declared in the bottom of this file.
"""
import os.path as op

from   nipype.pipeline.engine import Node, MapNode, JoinNode
from   nipype.interfaces.base import isdefined
from   kaptan import Kaptan


def _load_config(file_path):
    cpt = Kaptan()
    return cpt.import_config(file_path)


def _check_file(file_path):
    fpath = op.abspath(op.expanduser(file_path))

    if not op.isfile(fpath):
        raise IOError("Could not find configuration file {}.".format(fpath))


def _update_kaptan(kptn, new_values):
    for k, v in new_values.items():
        kptn.upsert(k, v)


class Borg:
    __shared_state = {}

    def __init__(self):
        self.__dict__ = self.__shared_state
    # and whatever else you want in your class -- that's all!


class Config(object):
    """ This class has a shared state like a Borg.

    This is a Kaptan class that infers the file handler
    from the file extension using the `from_file` function.

    Parameters
    ----------
    handler: str or kaptan.BaseHandler
        This parameter is to keep compatibility with Kaptan's instantiation protocol.
        See its documentation on how to use it: http://emre.github.io/kaptan/
        I personally prefer to use the `from_file` function.
    """
    __shared_state = {}

    def __init__(self, handler=None):
        self.__dict__ = self.__shared_state
        self._cpt = Kaptan(handler)

    @classmethod
    def from_file(cls, file_path):
        """ Returns a Config instance with the data from
        `file_path`.

        Parameters
        ----------
        file_path: str
            Path to a configuration file.
            Its extension can be any from config.HANDLER_EXT, i.e.,
            {'conf': 'ini',
             'ini': 'ini',
             'json': 'json',
             'py': 'file',
             'yaml': 'yaml',
             'yml': 'yaml'}

        Returns
        -------
        cfg: Config
        """
        _check_file(file_path)
        cfg = Config()
        cfg._cpt = _load_config(file_path)
        return cfg

    def update_from_file(self, file_path):
        """ Updates the config parameters with the data from
        `file_path`.

        Parameters
        ----------
        file_path: str
            Path to a configuration file.
            Its extension can be any from config.HANDLER_EXT, i.e.,
            {'conf': 'ini',
             'ini': 'ini',
             'json': 'json',
             'py': 'file',
             'yaml': 'yaml',
             'yml': 'yaml'}

        Returns
        -------
        cfg: Config
        """
        _check_file(file_path)
        cpt = _load_config(file_path)

        params = cpt.configuration_data
        _update_kaptan(self._cpt, params)

    def check_file(self, item):
        """ This is a __getitem__ operator for file path values, if the file does not exist an IOError is raised."""
        try:
            fpath = self.__getitem__(item)
        except KeyError:
            raise
        else:
            if not op.exists(fpath):
                raise IOError('Could not find file for key {} in the {}.'.format(item, fpath))
            return fpath

    def update(self, adict):
        for k, v in adict.items():
            self._cpt.upsert(k, v)

    def keys(self):
        return self._cpt.configuration_data.keys()

    def items(self):
        return self._cpt.configuration_data.items()

    def get(self, item, default=None):
        return self._cpt.configuration_data.get(item, default)

    def __getitem__(self, item):
        if item not in self._cpt.configuration_data:
            raise KeyError('Could not find key {} in configuration content.'.format(item))
        return self._cpt.configuration_data[item]

    def __setitem__(self, key, value):
        return self._cpt.upsert(key, value)

    def __repr__(self):
        return '<config.Config> ({})'.format('\n'.join([str(i) for i in self.items()]))


# the global configuration registry
PYPES_CFG = Config()


def node_settings(node_name):
    global PYPES_CFG
    for k, v in PYPES_CFG.items():
        if k.startswith(node_name):
            yield k, v


def update_config(value):
    """ Value can be a configuration file path or a dictionary with
    configuration settings."""
    global PYPES_CFG
    if isinstance(value, str):
        PYPES_CFG.update_from_file(value)
    elif isinstance(value, dict):
        PYPES_CFG.update(value)
    else:
        raise NotImplementedError('Cannot update the configuration with {}.'.format(value))


def _set_node_inputs(node, params, overwrite=False):
    for k, v in params.items():
        try:
            if not isdefined(getattr(node.inputs, k)):
                setattr(node.inputs, k, v)
            else:
                if overwrite:
                    setattr(node.inputs, k, v)
        except AttributeError as ate:
            raise AttributeError('Error in configuration settings: node `{}` '
                                 'has no attribute `{}`.'.format(node, k)) from ate


def _get_params_for(node_name):
    pars = {}
    for k, v in node_settings(node_name):
        nuk = '.'.join(k.split('.')[1:]) if '.' in k else k
        pars[nuk] = v

    return pars


def check_mandatory_inputs(node_names):
    """ Raise an exception if any of the items in the List[str] `node_names` is not
    present in the global configuration settings."""
    for name in node_names:
        if name not in PYPES_CFG:
            raise AttributeError('Could not find a configuration parameter for {}. '
                                 'Please set it in the an input configuration file.'.format(name))


def get_config_setting(param_name, default=''):
    """ Return the value for the entry with name `param_name` in the global configuration."""
    return PYPES_CFG.get(param_name, default=default)


def setup_node(interface, name, settings=None, overwrite=True, **kwargs):
    """ Create a pe.Node from `interface` with a given name.
    Check in the global configuration if there is any value for the node name and will set it.

    Parameters
    ----------
    interface: nipype.interface

    name: str

    settings: dict
        Dictionary with values for the pe.Node inputs.
        These will have higher priority than the ones in the global Configuration.

    overwrite: bool
        If True will overwrite the settings of the node if they are already defined.
        Default: True

    kwargs: keyword arguments
        type: str or None.
            choices: 'map', 'join, or None.
            If 'map' will return a MapNode.
            If 'join' will return a JoinNode.
            If None will return a Node.

        Extra arguments to pass to nipype.Node __init__ function.


    Returns
    -------
    node: nipype.Node
    """
    typ = kwargs.pop('type', None)
    if typ == 'map':
        node_class = MapNode
    elif typ == 'join':
        node_class = JoinNode
    else:
        node_class = Node
    node = node_class(interface=interface, name=name, **kwargs)

    params = _get_params_for(name)
    if settings is not None:
        params.update(settings)

    _set_node_inputs(node, params, overwrite=overwrite)

    return node


# ---------------------------------------------------------------------------
# Helper functions for specific parameters of config
# ---------------------------------------------------------------------------
def check_atlas_file():
    """ Return True and the path to the atlas_file if the configuration settings
    `normalize_atlas` is True and `atlas_file` points to an existing file.
    If `normalize_atlas` is False will return False and an empty string.
    Otherwise will raise a FileNotFoundError.

    Returns
    -------
    do_atlas: bool

    atlas_file: str
        Existing file path to the atlas file

    Raises
    ------
    FileNotFoundError
        If the `normalize_atlas` option is True but the
        `atlas_file` is not an existing file.
    """
    normalize_atlas = get_config_setting('normalize_atlas', default=False)
    if not normalize_atlas:
        return False, ''

    atlas_file = get_config_setting('atlas_file', default='')
    if not op.isfile(atlas_file):
        raise FileNotFoundError('Could not find atlas file in {}. '
                                'Please set `normalize_atlas` to False '
                                'or give an existing atlas image.'.format(atlas_file))
    return True, atlas_file




# -*- coding: utf-8 -*-
"""
A PETPVC workflow
"""
import os.path as op

import nipype.pipeline.engine as pe
from   nipype.algorithms.misc import Gunzip
from   nipype.interfaces.utility import Select, IdentityInterface, Function

from  .utils     import (petpvc_cmd,
                         petpvc_mask,
                         intensity_norm)
from   ..config  import setup_node, get_config_setting
from   ..preproc import spm_coregister
from   ..utils   import (get_datasink,
                         extend_trait_list,
                         get_input_node,
                         remove_ext,
                         get_input_file_name,
                         extension_duplicates)

from   .._utils import (flatten_list,
                        format_pair_list)


def petpvc_workflow(wf_name="petpvc"):
    """ Run the PET pre-processing workflow against the gunzip_pet.in_file files.
    It coregisters the reference_file and tissues to PET space, then applies PVC and grey matter normalization.

    It does:
    - SPM12 Coregister T1 and tisues to PET
    - PVC the PET image in PET space

    Parameters
    ----------
    wf_name: str
        Name of the workflow.

    Nipype Inputs
    -------------
    pvc_input.in_file: traits.File
        The raw NIFTI_GZ PET image file

    pvc_input.reference_file: traits.File
        The anatomical image in its native space. For registration reference.

    pvc_input.tissues: list of traits.File
        List of tissues files from the New Segment process. At least the first
        3 tissues must be present.

    Nipype outputs
    --------------
    pvc_output.coreg_ref: existing file
        The coregistered reference_file image in PET space.

    pvc_output.coreg_others: list of existing files
        List of coregistered files from coreg_pet.apply_to_files

    pvc_output.pvc_out: existing file
        The output of the PETPVC process.

    pvc_output.petpvc_mask: existing file
        The mask built for the PETPVC.

    pvc_output.brain_mask: existing file
        A brain mask calculated with the tissues file.

    pvc_output.gm_norm: existing file
        The output of the grey matter intensity normalization process.
        This is the last step in the PET signal correction.

    Returns
    -------
    wf: nipype Workflow
    """
    # fixed parameters of the NUK mMR
    psf_fwhm = (4.3, 4.3, 4.3)

    # specify input and output fields
    in_fields  = ["in_file",
                  "reference_file",
                  "tissues",]

    out_fields = ["coreg_ref",
                  "coreg_others",
                  "pvc_out",
                  "petpvc_mask",
                  "brain_mask",
                  "gm_norm",]

    # input
    pet_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                           name="pvc_input")

    flat_list = pe.Node(Function(input_names=['list_of_lists'], output_names=['out'],
                                 function=flatten_list),
                        name='flatten_tissue_list')

    # coreg pet
    gunzip_pet  = setup_node(Gunzip(),                           name="gunzip_pet")
    coreg_pet   = setup_node(spm_coregister(cost_function="mi"), name="coreg_pet")

    tissues_sel = setup_node(Select(index=[0, 1, 2]),            name="tissues")
    select_gm   = setup_node(Select(index=[0]),                  name="select_gm")
    rbvpvc      = setup_node(petpvc_cmd(fwhm_mm=psf_fwhm,
                                        pvc_method='RBV'),       name="rbvpvc")

    # output
    pvc_output = setup_node(IdentityInterface(fields=out_fields), name="pvc_output")

    # workflow to create the mask
    mask_wf = petpvc_mask(wf_name="petpvc_mask")

    # workflow for intensity normalization
    norm_wf = intensity_norm(wf_name="intensity_norm_gm")

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    wf.connect([
                # inputs
                (pet_input,   gunzip_pet,  [("in_file",  "in_file")]),
                (pet_input,   tissues_sel, [("tissues",  "inlist")]),
               ])

    # check how to perform the registration, to decide how to build the pipeline
    anat2pet = get_config_setting('registration.anat2pet', False)
    if anat2pet:
        wf.connect([
                    # inputs
                    (pet_input,   coreg_pet,   [("reference_file", "source")]),

                    # unzip to coregister the reference file (anatomical image) to PET space.
                    (gunzip_pet,  coreg_pet,  [("out_file", "target")]),

                    (tissues_sel, flat_list,  [("out", "list_of_lists")]),
                    (flat_list,   coreg_pet,  [("out", "apply_to_files")]),

                    # the list of tissues to the mask wf and the GM for PET intensity normalization
                    (coreg_pet,   select_gm,  [("coregistered_files", "inlist")]),
                    (coreg_pet,   mask_wf,    [("coregistered_files", "pvcmask_input.tissues")]),

                    # the PET in native space to PVC correction
                    (gunzip_pet,  rbvpvc,     [("out_file", "in_file")]),

                    # the merged file with 4 tissues to PCV correction
                    (mask_wf,     rbvpvc,     [("pvcmask_output.petpvc_mask", "mask_file")]),

                    # normalize voxel values of PET PVCed by demeaning it entirely by GM PET voxel values
                    (rbvpvc,      norm_wf,    [("out_file", "intnorm_input.source")]),
                    (select_gm,   norm_wf,    [("out",      "intnorm_input.mask")]),

                    # output
                    (coreg_pet,   pvc_output, [("coregistered_source",        "coreg_ref")]),
                    (coreg_pet,   pvc_output, [("coregistered_files",         "coreg_others")]),
                    (rbvpvc,      pvc_output, [("out_file",                   "pvc_out")]),
                    (mask_wf,     pvc_output, [("pvcmask_output.brain_mask",  "brain_mask")]),
                    (mask_wf,     pvc_output, [("pvcmask_output.petpvc_mask", "petpvc_mask")]),
                    (norm_wf,     pvc_output, [("intnorm_output.out_file",    "gm_norm")]),
                   ])
    else: # PET to ANAT
        wf.connect([
                    # inputs
                    (pet_input,   coreg_pet,   [("reference_file",     "target")]),

                    # unzip PET image and set as a source to register it to anatomical space.
                    (gunzip_pet,  coreg_pet,  [("out_file",            "source")]),

                    (tissues_sel, flat_list,  [("out", "list_of_lists")]),
                    (flat_list,   coreg_pet,  [("out", "apply_to_files")]),

                    # the list of tissues to the mask wf and the GM for PET intensity normalization
                    (tissues_sel, select_gm,  [("out", "inlist")]),
                    (flat_list,   mask_wf,    [("out", "pvcmask_input.tissues")]),

                    # the PET in ANAT space to PVC correction
                    (coreg_pet,    rbvpvc,     [("coregistered_source", "in_file")]),

                    # the merged file with 4 tissues to PCV correction
                    (mask_wf,     rbvpvc,     [("pvcmask_output.petpvc_mask", "mask_file")]),

                    # normalize voxel values of PET PVCed by demeaning it entirely by GM PET voxel values
                    (rbvpvc,      norm_wf,    [("out_file", "intnorm_input.source")]),
                    (select_gm,   norm_wf,    [("out",      "intnorm_input.mask")]),

                    # output
                    # TODO: coreg_ref should have a different name in this case
                    (coreg_pet,   pvc_output, [("coregistered_source",         "coreg_ref")]),
                    (coreg_pet,   pvc_output, [("coregistered_files",          "coreg_others")]),
                    (rbvpvc,      pvc_output, [("out_file",                    "pvc_out")]),
                    (mask_wf,     pvc_output, [("pvcmask_output.brain_mask",   "brain_mask")]),
                    (mask_wf,     pvc_output, [("pvcmask_output.petpvc_mask",  "petpvc_mask")]),
                    (norm_wf,     pvc_output, [("intnorm_output.out_file",     "gm_norm")]),
                   ])

    return wf


def attach_petpvc_workflow(main_wf, wf_name="spm_petpvc"):
    """ Attach a PETPVC workflow.

    This will also attach the anat preprocessing workflow to `main_wf`. The reason
    for this is that the PET pre-processing steps here make use of anatomical MR
    pre-processing outputs.

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.select.pet: input node

    datasink: nipype Node

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    Nipype Workflow Dependencies
    ----------------------------
    This workflow depends on:
    - spm_anat_preproc

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    anat_wf  = main_wf.get_node("spm_anat_preproc")
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    # The base name of the 'pet' file for the substitutions
    anat_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'anat')))
    pet_fbasename  = remove_ext(op.basename(get_input_file_name(in_files, 'pet')))

    # get the PET preprocessing pipeline
    pet_wf = petpvc_workflow(wf_name=wf_name)

    # dataSink output substitutions
    regexp_subst = [
                     (r"/{pet}_.*_pvc.nii.gz$",       "/{pet}_pvc.nii.gz"),
                     (r"/{pet}_.*_pvc_maths.nii.gz$", "/{pet}_pvc_norm.nii.gz"),
                     (r"/rm{anat}_corrected.nii$",    "/{anat}_{pet}.nii"),
                     (r"/rc1{anat}_corrected.nii$",   "/gm_{pet}.nii"),
                     (r"/rc2{anat}_corrected.nii$",   "/wm_{pet}.nii"),
                     (r"/rc3{anat}_corrected.nii$",   "/csf_{pet}.nii"),
                     (r"/tissues_brain_mask.nii$",    "/brain_mask_anat.nii.gz"),
                   ]
    regexp_subst = format_pair_list(regexp_subst, pet=pet_fbasename, anat=anat_fbasename)
    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    # Connect the nodes
    main_wf.connect([
                     # pet file input
                     (in_files, pet_wf, [("pet", "pvc_input.in_file")]),

                     # pet to anat registration
                     (anat_wf,  pet_wf, [("new_segment.bias_corrected_images", "pet_input.reference_file"),
                                         ("new_segment.native_class_images",   "pet_input.tissues"),
                                        ]),

                     (pet_wf, datasink, [
                                         ("pvc_output.coreg_others", "mrpet.tissues"),
                                         ("pvc_output.coreg_ref",    "mrpet.@anat"),
                                         ("pvc_output.pvc_out",      "mrpet.@pvc"),
                                         ("pvc_output.petpvc_mask",  "mrpet.@petpvc_mask"),
                                         ("pvc_output.brain_mask",   "mrpet.@brain_mask"),
                                         ("pvc_output.gm_norm",      "mrpet.@gm_norm"),
                                        ]),
                     ])

    return main_wf


def merge(output_dir, scan_name, threshold, motion_f, power_f, flag):

    """
    Method to merge power parameters and motion
    parameters file
    """

    import os
    import re

    if threshold == None:
        filename = scan_name + "_all_params.csv"
        filename = filename.lstrip("_")
        outfile = os.path.join(output_dir, filename)

        threshold_val = 0.0

    else:
        filename = scan_name + threshold + "_all_params.csv"
        filename = filename.lstrip("_")
        outfile = os.path.join(output_dir, filename)

        threshold_val = float(re.sub(r"[a-zA-Z_]", '', threshold))


    # Read in the motion and power parameters files
    try:
        motion = open(motion_f, 'r').readlines()
    except Exception as e:
        err_string = "\n\n[!] CPAC says: Could not read the motion " \
                     "parameters file.\n\nFilepath: %s\n\nError details: %s" \
                     "\n\n" % (motion_f, e)
        raise Exception(err_string)

    try:
        power = open(power_f, 'r').readlines()
    except Exception as e:
        err_string = "\n\n[!] CPAC says: Could not read the power " \
                     "parameters file.\n\nFilepath: %s\n\nError details: %s" \
                     "\n\n" % (power_f, e)
        raise Exception(err_string)


    # Write the combined motion and power parameters CSV file
    try:

        if flag:
            f = open(outfile, 'w')

            print "Combining motion and power parameters into one CSV file: " \
                  "%s\n" % outfile
            print >>f, "Subject,Scan,Mean_Relative_RMS_Displacement," \
            "Max_Relative_RMS_Displacement,Movements_gt_threshold,"\
            "Mean_Relative_Mean_Rotation,Mean_Relative_Maxdisp,Max_Relative_Maxdisp," \
            "Max_Abs_Maxdisp,Max_Relative_Roll,Max_Relative_Pitch," \
            "Max_Relative_Yaw,Max_Relative_dS-I,Max_Relative_dL-R," \
            "Max_Relative_dP-A,Mean_Relative_Roll,Mean_Relative_Pitch,Mean_Relative_Yaw," \
            "Mean_Relative_dS-I,Mean_Relative_dL-R,Mean_Relative_dP-A,Max_Abs_Roll," \
            "Max_Abs_Pitch,Max_Abs_Yaw,Max_Abs_dS-I,Max_Abs_dL-R,Max_Abs_dP-A," \
            "Mean_Abs_Roll,Mean_Abs_Pitch,Mean_Abs_Yaw,Mean_Abs_dS-I,Mean_Abs_dL-R,Mean_Abs_dP-A,"\
            "MeanFD,NumFD_greater_than_%.2f,rootMeanSquareFD,FDquartile(top1/4thFD),"\
            "PercentFD_greater_than_%.2f,MeanDVARS,MeanFD_Jenkinson" % (threshold_val, threshold_val)
        else:
             f = open(outfile, 'a')

        m = motion[1].rstrip(",").split(",")
        p = power[1].rstrip(",").split(",")

        for item in m:
            f.write("%s," % item)

        for item in p[2:]:
            f.write("%s," % item)
        f.write("\n")
        f.close()

    except Exception as e:

        err_string = "\n\n[!] CPAC says: Could not create or open the motion "\
                     "and power parameters CSV file. Ensure you have write " \
                     "permissions for the directory it is writing to.\n\n" \
                     "Attempted write path: %s\n\nError details: %s\n\n" \
                     % (outfile, e)

        raise Exception(err_string)



def grab(output_dir, scrubbing):

    """
    Method to grab all the motion parameters
    and power parameters file from each subject
    for each pipeline and merge them

    Parameters
    ----------
    output_dir : string
        Path to the datasink output directory of CPAC
    """

    import glob
    import os
    import re
    from sets import Set

    pipelines = glob.glob(os.path.join(output_dir, 'pipeline*'))

    print "number of pipelines ", len(pipelines)

    for p in pipelines:
        print "inside pipeline ->", os.path.basename(p)
        scan_list = []
        threshold_list = []

        pattern1 = re.compile(r'(\w)*scan(\w)*(\d)*(\w)*[/]')
        pattern2 = re.compile(r'(\w)*threshold_[-+]?([0-9]*\.[0-9]+|[0-9]+)')

        scans = glob.glob(os.path.join(p, '*/power_params/*/*'))

        #get the unique scans and threshold value
        for s in scans:
            val = re.search(pattern1, s)
            if val:
                scan_list.append(val.group(0).rstrip("/"))

            val = re.search(pattern2, s)
            if val:
                threshold_list.append(val.group(0))

        scan_list = Set(scan_list)
        threshold_list = Set(threshold_list)

        #print "scan_list ->", scan_list
        #print "threshold_list ->", threshold_list

        for scan in scan_list:
            for threshold in threshold_list:
                print "running for...", scan, threshold
                Flag = 1
                #merge files for each subject
                for sub in os.listdir(p):
                    sub = os.path.join(p, sub)
                    motion_file = os.path.join(sub, 'motion_params', scan,
                                               'motion_parameters.txt')
                    power_file = os.path.join(sub, 'power_params', scan,
                                              threshold, 'pow_params.txt')

                    if os.path.exists(motion_file) and \
                        os.path.exists(power_file):
                        merge(p, scan, threshold,
                              motion_file, power_file, Flag)
                        Flag = 0


            if 0 in scrubbing:
                for sub in os.listdir(p):
                    sub = os.path.join(p, sub)
                    motion_file = os.path.join(sub, 'motion_params', scan,
                                               'motion_parameters.txt')
                    power_file = os.path.join(sub, 'power_params', scan,
                                              'pow_params.txt')

                    if os.path.exists(motion_file) and \
                        os.path.exists(power_file):

                        threshold = None

                        merge(p, scan, threshold, motion_file,
                            power_file, Flag)

                        Flag = 0
                

    print "Motion and Power parameters extraction process finished...\n"

    return threshold

    

def run(output_path, scrubbing):
    threshold = grab(output_path, scrubbing)
    return threshold


def main():
    import sys
    if (len(sys.argv) == 2):
        grab(sys.argv[1])
    else:
        print 'Usage: cpac_extract_parameters /path/to/datasink_dir'



# -*- coding: utf-8 -*-
"""
Diffusion Tensor Image preprocessing and tensor model fitting workflow.
"""
import os.path as op

from .artifacts import attach_dti_artifact_correction
from .coregister import spm_anat_to_diff_coregistration
from .._utils  import format_pair_list
from ..config  import check_atlas_file
from ..utils   import (get_datasink,
                       get_input_node,
                       get_interface_node,
                       remove_ext,
                       extend_trait_list,
                       get_input_file_name,
                       extension_duplicates,
                       )


def attach_spm_fsl_dti_preprocessing(main_wf, wf_name="spm_fsl_dti_preprocessing"):
    """ Attach a set of pipelines to the `main_wf` for Diffusion MR (`diff`) image processing.
    - dti_artifact_correction
    - spm_anat_to_diff_coregistration
    - dti_tensor_fitting

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    params: dict with parameter values
        atlas_file: str
            Path to the anatomical atlas to be transformed to diffusion MRI space.

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.select.diff: input node

    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)
    anat_output = get_interface_node(main_wf, 'anat_output')

    # attach the artifact detection and correction pipeline
    main_wf = attach_dti_artifact_correction(main_wf)
    dti_art_output = get_interface_node(main_wf, 'dti_art_output')

    # The workflow boxes
    coreg_dti_wf = spm_anat_to_diff_coregistration(wf_name=wf_name)

    # dataSink output substitutions
    ## The base name of the 'diff' file for the substitutions
    diff_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'diff')))
    anat_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'anat')))

    regexp_subst = [
                    (r"/brain_mask_{diff}_space\.nii$", "/brain_mask.nii"),
                    (r"/eddy_corrected\.nii$",          "/{diff}_eddycor.nii"),
                    (r"/rc1anat_hc_corrected\.nii$",    "/gm_diff.nii"),
                    (r"/rc2anat_hc_corrected\.nii$",    "/wm_diff.nii"),
                    (r"/rc3anat_hc_corrected\.nii$",    "/csf_diff.nii"),
                    (r"/rmanat_hc_corrected\.nii$",     "/{anat}_diff.nii"),
                   ]
    regexp_subst = format_pair_list(regexp_subst, diff=diff_fbasename,
                                                  anat=anat_fbasename)

    # prepare substitution for atlas_file, if any
    do_atlas, atlas_file = check_atlas_file()
    if do_atlas:
        atlas_basename = remove_ext(op.basename(atlas_file))
        regexp_subst.extend([
                             (r"/[\w]*{atlas}.*\.nii$", "/{atlas}_{diff}_space.nii"),
                            ])
        regexp_subst = format_pair_list(regexp_subst, atlas=atlas_basename,
                                                      diff=diff_fbasename)


    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    # input and output diffusion MRI workflow to main workflow connections
    main_wf.connect([
                     (dti_art_output, coreg_dti_wf, [("avg_b0",         "dti_co_input.avg_b0"),]),
                     (anat_output,    coreg_dti_wf, [("tissues_native", "dti_co_input.tissues"),
                                                     ("anat_biascorr",  "dti_co_input.anat")
                                                    ]),
                     (coreg_dti_wf, datasink, [("dti_co_output.anat_diff",       "diff.@anat_diff"),
                                               ("dti_co_output.tissues_diff",    "diff.tissues.@tissues_diff"),
                                               ("dti_co_output.brain_mask_diff", "diff.@brain_mask"),
                                              ]),
                    ])

    if do_atlas:
            main_wf.connect([(anat_output,  coreg_dti_wf, [("atlas_anat", "dti_co_input.atlas_anat")]),
                             (coreg_dti_wf, datasink,     [("dti_co_output.atlas_diff", "diff.@atlas")]),
                            ])

    return main_wf





# -*- coding: utf-8 -*-
"""
Helper functions to read, threshold, and build IC loadings results.
"""
import pandas as pd
import nilearn.image as niimg
from   nilearn.image import iter_img
from   boyle.nifti.roi import largest_connected_component
from   boyle.nifti.utils import filter_icc


def get_largest_blobs(ic_maps):
    """ Generator for the largest blobs in each IC spatial map.
    These should be masked and thresholded.

    Parameters
    ----------
    ic_maps: sequence of niimg-like

    Returns
    -------
    blobs: generator of niimg-like
    """
    # store the average value of the blob in a list
    for i, icimg in enumerate(iter_img(ic_maps)):
        yield niimg.new_img_like(icimg, largest_connected_component(icimg.get_data()))


def build_raw_loadings_table(loads, patids):
    """ Build a spreadsheet-ready pandas.DataFrame with the content of the
    loadings matrix and subjects ids.
    """
    loadings = []
    for p in range(len(patids)):
        patid = patids[p]
        loadings.append([patid] + list(loads[p, :]))

    # set the column names
    n_cols = loads.shape[1]
    cols = ['subject_id'] + list(range(1, n_cols+1))

    # fill the df
    return pd.DataFrame.from_records(loadings, columns=cols)


def add_groups_to_loadings_table(df, groups):
    """ Merge `df` and `groups` on 'subject_id' and sort by 'group'.
    Note: `groups` must have 'subject_id' (as well as `df`) and 'group' columns.
    """
    if groups is not None:
        df = pd.merge(df, groups, on='subject_id')
        df = df.sort_values(by=['group', 'subject_id'], ascending=[True, True])

    df.loc[:, 'group'] = df.loc[:, 'group'].astype('category')
    return df


def filter_ics(comps_img, mask, zscore=2., mode='+-'):
    """
    Generator for masking and thresholding each IC spatial map.

    Parameters
    ----------
    comps_img: img-like
        The 'raw' ICC maps image.

    mask: img-like
        If not None. Will apply this masks in the end of the process.

    thr: float
        The threshold value.

    zscore: bool
        If True will calculate the z-score of the ICC before thresholding.

    mode: str
        Choices: '+' for positive threshold,
                 '+-' for positive and negative threshold and
                 '-' for negative threshold.

    Returns
    -------
    icc_filts: list of nibabel.NiftiImage
        Thresholded and masked ICCs.
    """
    # store the average value of the blob in a list
    mask = niimg.load_img(mask)
    for i, icimg in enumerate(iter_img(comps_img)):
        yield filter_icc(icimg, mask=mask, thr=zscore, zscore=True, mode=mode)



# -*- coding: utf-8 -*-
"""
Nipype workflows to process anatomical MRI.
"""
import os.path as op

import nipype.pipeline.engine    as pe
from   nipype.algorithms.misc    import Gunzip
from   nipype.interfaces.utility import IdentityInterface, Function
from   nipype.interfaces.base    import isdefined

from .utils import biasfield_correct, spm_segment

from   ..interfaces.nilearn import math_img, copy_header
from   ..config  import setup_node, check_atlas_file, get_config_setting
from   ..preproc import (spm_apply_deformations,
                         get_bounding_box,)
from   .._utils  import format_pair_list
from   ..utils   import (remove_ext,
                         selectindex,
                         spm_tpm_priors_path,
                         extend_trait_list,
                         get_input_node,
                         get_datasink,
                         get_input_file_name,
                         extension_duplicates)


def spm_anat_preprocessing(wf_name="spm_anat_preproc"):
    """ Run the T1 pre-processing workflow against the anat_hc
    files in `data_dir`.

    It does:
    - N4BiasFieldCorrection
    - SPM12 New Segment
    - SPM12 Warp of MPRAGE to MNI

    [Optional: from config]
    - Atlas file warping to MPRAGE
    - Cortical thickness (SPM+DiReCT)

    Nipype Inputs
    -------------
    anat_input.in_file: traits.File
        Path to the anatomical image.

    anat_input.atlas_file: traits.File
        Path to an atlas file in MNI space to be
        warped to the anatomical space.
        Can also be set through the configuration
        setting `atlas_file`.

    Nipype Outputs
    --------------
    anat_output.anat_mni: traits.File
        The bias-field normalized to MNI anatomical image.

    anat_output.tissues_warped: traits.File
        The tissue segmentation in MNI space from SPM.

    anat_output.tissues_native: traits.File
        The tissue segmentation in native space from SPM.

    anat_output.affine_transform: traits.File
        The affine transformation file.

    anat_output.warp_forward: traits.File
        The forward (anat to MNI) warp field from SPM.

    anat_output.warp_inverse: traits.File
        The inverse (MNI to anat) warp field from SPM.

    anat_output.anat_biascorr: traits.File
        The bias-field corrected anatomical image.

    anat_output.brain_mask: traits.File
        A brain mask file in anatomical space.
        This is calculated by summing up the maps of
        segmented tissues (CSF, WM, GM) and then binarised.

    anat_output.atlas_anat: traits.File
        If `atlas_file` is an existing file in MNI space.
        The atlas file warped to anatomical space,
        if do_atlas and the atlas file is set in configuration.

    anat_output.cortical_thickness: traits.File
        If `anat_preproc.do_cortical_thickness` is True.
        The cortical thickness estimations calculated with the
        SPM+DiReCT method (KellyKapowski).

    anat_output.warped_white_matter: warped_white_matter
        If `anat_preproc.do_cortical_thickness` is True.
        The warped white matter image calculated with the
        SPM+DiReCT method (KellyKapowski).

    Returns
    -------
    wf: nipype Workflow
    """
    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # specify input and output fields
    in_fields  = ["in_file"]
    out_fields = ["anat_mni",
                  "tissues_warped",
                  "tissues_native",
                  "affine_transform",
                  "warp_forward",
                  "warp_inverse",
                  "anat_biascorr",
                  "brain_mask",
                 ]

    # check if we have to warp an atlas files too.
    do_atlas, atlas_file = check_atlas_file()
    if do_atlas:
        in_fields  += ["atlas_file"]
        out_fields += ["atlas_anat"]

    # check if we have to do cortical thickness (SPM+DiReCT) method.
    do_cortical_thickness = get_config_setting('anat_preproc.do_cortical_thickness', False)
    if do_cortical_thickness:
        out_fields += ["cortical_thickness",
                       "warped_white_matter",]

    # input node
    anat_input = pe.Node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                         name="anat_input")

    # atlas registration
    if do_atlas and not isdefined(anat_input.inputs.atlas_file):
        anat_input.inputs.set(atlas_file=atlas_file)

    # T1 preprocessing nodes
    biascor     = setup_node(biasfield_correct(),      name="bias_correction")
    gunzip_anat = setup_node(Gunzip(),                 name="gunzip_anat")
    segment     = setup_node(spm_segment(),            name="new_segment")
    warp_anat   = setup_node(spm_apply_deformations(), name="warp_anat")

    tpm_bbox = setup_node(Function(function=get_bounding_box,
                                   input_names=["in_file"],
                                   output_names=["bbox"]),
                          name="tpm_bbox")
    tpm_bbox.inputs.in_file = spm_tpm_priors_path()

    # calculate brain mask from tissue maps
    tissues = setup_node(IdentityInterface(fields=["gm", "wm", "csf"], mandatory_inputs=True),
                         name="tissues")

    brain_mask = setup_node(Function(function=math_img,
                                     input_names=["formula", "out_file", "gm", "wm", "csf"],
                                     output_names=["out_file"],
                                     imports=['from neuro_neuro_pypes.interfaces.nilearn import ni2file']),
                            name='brain_mask')
    brain_mask.inputs.out_file = "tissues_brain_mask.nii.gz"
    brain_mask.inputs.formula  = "np.abs(gm + wm + csf) > 0"

    # output node
    anat_output = pe.Node(IdentityInterface(fields=out_fields), name="anat_output")

    # Connect the nodes
    wf.connect([
                # input to biasfieldcorrection
                (anat_input,   biascor    , [("in_file",      "input_image")]),

                # new segment
                (biascor,      gunzip_anat, [("output_image", "in_file")]),
                (gunzip_anat,  segment,     [("out_file",     "channel_files")]),

                # Normalize12
                (segment,   warp_anat,  [("forward_deformation_field", "deformation_file")]),
                (segment,   warp_anat,  [("bias_corrected_images",     "apply_to_files")]),
                (tpm_bbox,  warp_anat,  [("bbox",                      "write_bounding_box")]),

                # brain mask from tissues
                (segment, tissues,  [(("native_class_images", selectindex, 0), "gm"),
                                     (("native_class_images", selectindex, 1), "wm"),
                                     (("native_class_images", selectindex, 2), "csf"),
                                    ]),

                (tissues, brain_mask, [("gm", "gm"), ("wm", "wm"), ("csf", "csf"),]),

                # output
                (warp_anat, anat_output, [("normalized_files",           "anat_mni")]),
                (segment,   anat_output, [("modulated_class_images",     "tissues_warped"),
                                          ("native_class_images",        "tissues_native"),
                                          ("transformation_mat",         "affine_transform"),
                                          ("forward_deformation_field",  "warp_forward"),
                                          ("inverse_deformation_field",  "warp_inverse"),
                                          ("bias_corrected_images",      "anat_biascorr")]),
                (brain_mask, anat_output, [("out_file",                  "brain_mask")]),
              ])

    # atlas warping nodes
    if do_atlas:
        gunzip_atlas = pe.Node(Gunzip(), name="gunzip_atlas")
        warp_atlas   = setup_node(spm_apply_deformations(), name="warp_atlas")
        anat_bbox    = setup_node(Function(function=get_bounding_box,
                                           input_names=["in_file"],
                                           output_names=["bbox"]),
                                  name="anat_bbox")

        # set the warping interpolation to nearest neighbour.
        warp_atlas.inputs.write_interp = 0

        # connect the atlas registration nodes
        wf.connect([
                    (anat_input,    gunzip_atlas, [("atlas_file",                 "in_file")]),
                    (gunzip_anat,   anat_bbox,    [("out_file",                   "in_file")]),
                    (gunzip_atlas,  warp_atlas,   [("out_file",                   "apply_to_files")]),
                    (segment,       warp_atlas,   [("inverse_deformation_field",  "deformation_file")]),
                    (anat_bbox,     warp_atlas,   [("bbox",                       "write_bounding_box")]),
                    (warp_atlas,    anat_output,  [("normalized_files",           "atlas_anat")]),
                  ])

    # cortical thickness (SPM+DiReCT) method
    if do_cortical_thickness:
        from ..interfaces.ants import KellyKapowski

        segm_img = setup_node(Function(function=math_img,
                                       input_names=["formula", "out_file", "gm", "wm"],
                                       output_names=["out_file"],
                                       imports=['from neuro_neuro_pypes.interfaces.nilearn import ni2file']),
                              name='gm-wm_image')
        segm_img.inputs.out_file = "gm_wm.nii.gz"
        segm_img.inputs.formula = '((gm >= 0.5)*2 + (wm > 0.5)*3).astype(np.uint8)'

        # copy the header from the GM tissue image to the result from `gm-wm_image`.
        # this is necessary because the `gm-wm_image` operation sometimes modifies the
        # offset of the image, which will provoke an ANTs exception due to
        # ITK tolerance in ImageToImageFilter
        # https://github.com/stnava/ANTs/issues/74
        cp_hdr = setup_node(Function(function=copy_header,
                                     input_names=["in_file", "data_file"],
                                     output_names=["out_file"],
                                     imports=['from neuro_neuro_pypes.interfaces.nilearn import ni2file']),
                            name='copy_header')

        kk = setup_node(KellyKapowski(), name='direct')
        kk.inputs.cortical_thickness  = 'direct_cortical_thickness.nii.gz'
        kk.inputs.warped_white_matter = 'direct_warped_white_matter.nii.gz'

        # connect the cortical thickness (SPM+DiReCT) method
        wf.connect([
                    # create segmentation GM+WM file
                    (tissues, segm_img, [("gm", "gm"),
                                         ("wm", "wm")]),

                    (segm_img, cp_hdr, [("out_file", "data_file")]),
                    (tissues,  cp_hdr, [("gm",       "in_file")]),

                    # kellykapowski
                    (cp_hdr,   kk, [("out_file", "segmentation_image")]),
                    (tissues,  kk, [("gm", "gray_matter_prob_image"),
                                    ("wm", "white_matter_prob_image")]),

                    (kk,    anat_output,  [("cortical_thickness",  "cortical_thickness"),
                                           ("warped_white_matter", "warped_white_matter")]),
                  ])
    return wf


def attach_spm_anat_preprocessing(main_wf, wf_name="spm_anat_preproc"):
    """ Attach the SPM12 anatomical MRI pre-processing workflow to
    the `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an
    `input_files` and a `datasink` nodes.

    input_files.anat: input node

    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    # The workflow box
    anat_wf = spm_anat_preprocessing(wf_name=wf_name)

    # The base name of the 'anat' file for the substitutions
    anat_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'anat')))

    # dataSink output substitutions
    regexp_subst = [
                    (r"/{anat}_.*corrected_seg8.mat$",    "/{anat}_to_mni_affine.mat"),
                    (r"/m{anat}.*_corrected.nii$",        "/{anat}_biascorrected.nii"),
                    (r"/wm{anat}.*_corrected.nii$",       "/{anat}_mni.nii"),
                    (r"/y_{anat}.*nii$",                  "/{anat}_to_mni_field.nii"),
                    (r"/iy_{anat}.*nii$",                 "/{anat}_to_mni_inv_field.nii"),
                    (r"/mwc1{anat}.*nii$",                "/{anat}_gm_mod_mni.nii"),
                    (r"/mwc2{anat}.*nii$",                "/{anat}_wm_mod_mni.nii"),
                    (r"/mwc3{anat}.*nii$",                "/{anat}_csf_mod_mni.nii"),
                    (r"/mwc4{anat}.*nii$",                "/{anat}_nobrain_mod_mni.nii"),
                    (r"/c1{anat}.*nii$",                  "/{anat}_gm.nii"),
                    (r"/c2{anat}.*nii$",                  "/{anat}_wm.nii"),
                    (r"/c3{anat}.*nii$",                  "/{anat}_csf.nii"),
                    (r"/c4{anat}.*nii$",                  "/{anat}_nobrain.nii"),
                    (r"/c5{anat}.*nii$",                  "/{anat}_nobrain_mask.nii"),
                    (r"/direct_cortical_thickness.nii$",  "/{anat}_gm_cortical_thickness.nii"),
                    (r"/direct_warped_white_matter.nii$", "/{anat}_warped_white_matter.nii"),
                   ]
    regexp_subst = format_pair_list(regexp_subst, anat=anat_fbasename)

    # prepare substitution for atlas_file, if any
    do_atlas, atlas_file = check_atlas_file()
    if do_atlas:
        atlas_basename = remove_ext(op.basename(atlas_file))
        regexp_subst.extend([
                             (r"/w{atlas}\.nii$", "/{atlas}_anat_space.nii"),
                            ])
        regexp_subst = format_pair_list(regexp_subst, atlas=atlas_basename)

    # add nii.gz patterns
    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    main_wf.connect([(in_files, anat_wf,  [("anat",                         "anat_input.in_file")]),
                     (anat_wf,  datasink, [("anat_output.anat_mni",         "anat.@mni"),
                                           ("anat_output.tissues_warped",   "anat.tissues.warped"),
                                           ("anat_output.tissues_native",   "anat.tissues.native"),
                                           ("anat_output.affine_transform", "anat.transform.@linear"),
                                           ("anat_output.warp_forward",     "anat.transform.@forward"),
                                           ("anat_output.warp_inverse",     "anat.transform.@inverse"),
                                           ("anat_output.anat_biascorr",    "anat.@biascor"),
                                           ("anat_output.brain_mask",       "anat.@brain_mask"),
                                          ]),
                    ])

    # check optional outputs
    if do_atlas:
        main_wf.connect([(anat_wf, datasink, [("anat_output.atlas_anat", "anat.@atlas")]),])

    do_cortical_thickness = get_config_setting('anat_preproc.do_cortical_thickness', False)
    if do_cortical_thickness:
        main_wf.connect([(anat_wf, datasink, [("anat_output.cortical_thickness",  "anat.@cortical_thickness"),
                                              ("anat_output.warped_white_matter", "anat.@warped_white_matter"),
                                             ]),
                        ])

    return main_wf




from .nilearn.canica import CanICAInterface

from .nilearn.plot import (plot_all_components,
                           plot_ica_components,
                           plot_multi_slices,
                           plot_ortho_slices,
                           plot_overlays,
                           plot_stat_overlay)

from .ants import KellyKapowski

from .fsl import EddyCorrect, Eddy





from .warp import attach_spm_warp_fmri_wf
from .clean import attach_fmri_cleanup_wf
from .rest import attach_rest_preprocessing, attach_rest_grptemplate_preprocessing
from .grouptemplate import attach_spm_fmri_grouptemplate_wf


# -*- coding: utf-8 -*-
"""
Plotting utilities to use nilearn from nipype
"""
import logging


def plot_all_components(components_img, **kwargs):
    """ Plot the components IC spatial maps in only one brain. """
    from nilearn.plotting import plot_prob_atlas
    from matplotlib import pyplot as plt

    fig = plt.figure(facecolor='white')
    p   = plot_prob_atlas(components_img, figure=fig, draw_cross=False, **kwargs)
    p.close()

    return fig


def plot_ica_components(components_img, **kwargs):
    """ Plot the components IC spatial maps in a grid."""
    import math
    from nilearn.image import iter_img
    from nilearn.plotting import plot_stat_map
    from matplotlib import pyplot as plt
    from matplotlib import gridspec

    n_ics  = len(list(iter_img(components_img)))
    n_rows = math.ceil(n_ics/2)
    fig = plt.figure(figsize=(6, 3*n_rows), facecolor='black')
    gs  = gridspec.GridSpec(n_rows, 2)

    plots = []
    for i, ic_img in enumerate(iter_img(components_img)):
        ax = plt.subplot(gs[i])
        p  = plot_stat_map(ic_img, display_mode="z", title="IC {}".format(i+1),
                           cut_coords=1, colorbar=False, figure=fig, axes=ax, **kwargs)
        plots.append(p)

    for p in plots:
        p.close()

    return fig


def plot_multi_slices(img, cut_dir="z", n_cuts=20, n_cols=4, figsize=(2.5, 3),
                      title="", title_fontsize=32, plot_func=None, black_bg=True, **kwargs):
    """ Create a plot of `n_cuts` of `img` organized distributed in `n_cols`.
    Parameters
    ----------
    img: niimg-like

    cut_dir: str, optional
        Sectional direction; possible values are "x", "y" or "z".

    n_cuts: int, optional
        Number of cuts in the plot.

    n_cols: int, optional
        Maximum number of image columns in the plot.

    figsize: 2-tuple of int, optional
        (w, h) size in inches of the figure.

    title: str, optional
        The superior title of the figure.

    title_fontsize: int, optional
        The size of the title font.

    plot_func: function, optional
       Function to plot each slice.
       Default: nilearn.plotting.plot_stat_map

    black_bg: boolean, optional
        If True, the background of the image is set to be black.
        If you wish to save figures with a black background, you will need to pass
        "facecolor='k', edgecolor='k'" to matplotlib.pyplot.savefig.

    kwargs: keyword arguments, optional
        Input arguments for plot_func.

    Returns
    -------
    fig: matplotlib.figure
    """
    import math

    import matplotlib.pyplot as plt
    from matplotlib import gridspec
    import nilearn.plotting as niplot
    import nilearn.image as niimg

    # there is another version without grouper, but it is less efficient.
    def grouper(iterable, n, fillvalue=None):
        "Collect data into fixed-length chunks or blocks"
        # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
        import itertools
        args = [iter(iterable)] * n
        return itertools.zip_longest(fillvalue=fillvalue, *args)

    if plot_func is None:
        plot_func = niplot.plot_stat_map

    _img = niimg.load_img(img)

    n_rows = 1
    if n_cuts > n_cols:
        n_rows = math.ceil(n_cuts/n_cols)

    spacing = kwargs.get('spacing', 'auto')
    cuts = niplot.find_cut_slices(_img, n_cuts=n_cuts,
                                  direction=cut_dir,
                                  spacing=spacing)

    # instantiate the figure
    if black_bg:
        facecolor = 'black'
        titlecolor = 'white'
    else:
        facecolor = 'white'
        titlecolor = 'black'

    figsize = figsize[0] * n_cols, figsize[1] * n_rows
    fig = plt.figure(figsize=figsize, facecolor=facecolor)
    gs  = gridspec.GridSpec(n_rows, 1)

    if title:
        fig.suptitle(title, fontsize=title_fontsize,
                     color=titlecolor)

    # for the superior plot
    put_colorbar = True

    # make the plots
    for i, cut_chunks in enumerate(grouper(cuts, n_cols)):
        ax = plt.subplot(gs[i])

        cut_chunks = [cut for cut in cut_chunks if cut is not None]

        try:
            p = plot_func(_img,
                          display_mode=cut_dir,
                          cut_coords=cut_chunks,
                          colorbar=put_colorbar,
                          figure=fig,
                          axes=ax,
                          black_bg=black_bg,
                          **kwargs)
        except IndexError:
            logging.warning('Could not plot for coords {}.'.format(cut_chunks))
        finally:
            put_colorbar = False

    return fig


def plot_ortho_slices(img, n_cuts=4, n_cols=6, figsize=(2.5, 3),
                      title="", title_fontsize=32, plot_func=None,
                      black_bg=True, **kwargs):
    """
    Parameters
    ----------
    img: niimg-like

    n_cuts: int, optional
        Number of cuts for each dimension (X, Y, and Z) in the plot.

    n_cols: int, optional
        Maximum number of image columns in the plot.

    figsize: 2-tuple of int, optional
        (w, h) size in inches of one slice image.

    title: str, optional
        The superior title of the figure.

    title_fontsize: int, optional
        The size of the title font.

    plot_func: function, optional
       Function to plot each slice.
       Default: nilearn.plotting.plot_stat_map

    black_bg: boolean, optional
        If True, the background of the image is set to be black.
        If you wish to save figures with a black background, you will need to pass
        "facecolor='k', edgecolor='k'" to matplotlib.pyplot.savefig.

    kwargs: keyword arguments, optional
        Input arguments for plot_func.

    Returns
    -------
    fig: matplotlib.figure
    """
    from matplotlib import pyplot as plt
    from matplotlib import gridspec
    import nilearn.plotting as niplot
    import nilearn.image as niimg

    if plot_func is None:
        plot_func = niplot.plot_stat_map

    # load the image file
    _img = niimg.load_img(img)

    directions = ('x', 'y', 'z')

    # calculate the shape of the figure
    total_cuts = n_cuts * 3
    if total_cuts > n_cols:
        n_rows = 3
        n_cols = 1
        colorbard_idx = 0
    else:
        n_rows = 1
        n_cols = 3
        colorbard_idx = len(directions)-1

    # calculate the cut coordinates for each direction
    cuts = []
    spacing = kwargs.get('spacing', 'auto')
    for cut_dir in directions:
        dir_cuts = niplot.find_cut_slices(_img, n_cuts=n_cuts,
                                          direction=cut_dir,
                                          spacing=spacing)
        cuts.append((cut_dir, dir_cuts))

    # instantiate the figure
    if black_bg:
        facecolor = 'black'
        titlecolor = 'white'
    else:
        facecolor = 'white'
        titlecolor = 'black'

    figsize = figsize[0] * n_cols * n_cuts, figsize[1] * n_rows
    fig = plt.figure(figsize=figsize, facecolor=facecolor)
    gs  = gridspec.GridSpec(n_rows, n_cols)

    # put the title, if any
    if title:
        fig.suptitle(title, fontsize=title_fontsize,
                     color=titlecolor)

    # plot on the figure
    for i, (cut_dir, cut_coords) in enumerate(cuts):
        ax = plt.subplot(gs[i])

        put_colorbar = True if i == colorbard_idx else False
        try:
            p = plot_func(_img,
                          display_mode=cut_dir,
                          cut_coords=cut_coords,
                          colorbar=put_colorbar,
                          figure=fig,
                          axes=ax,
                          black_bg=black_bg,
                          **kwargs)
        except IndexError:
            logging.warning('Could not plot for coords {}.'.format(cut_coords))

    return fig


def plot_stat_overlay(stat_img, contour_img, bg_img, **kwargs):
    """Plot over bg_img a stat_img and the countour."""
    import nilearn.plotting as niplot

    if bg_img is not None:
        kwargs['bg_img'] = bg_img

    display = niplot.plot_stat_map(stat_img, **kwargs)
    display.add_contours(contour_img, filled=True, alpha=0.6, levels=[0.5], colors='g')
    return display


def plot_overlays(stat_imgs, contour_imgs, bg_img=None,
                  figsize=(2.5, 3), title_fontsize=32, title='', **kwargs):
    """Plots each contour_imgs as an overlay of its corresponding `stat_imgs`.
    `contour_imgs` and `stat_imgs` must have the same length."""

    from matplotlib import pyplot as plt
    from matplotlib import gridspec

    _stat_imgs = list(stat_imgs)
    _cnts_imgs = list(contour_imgs)
    n_stats    = len(_stat_imgs)
    n_conts    = len(_cnts_imgs)
    if n_stats != n_conts:
        raise AttributeError('The length of `stat_imgs` and `contour_imgs` are '
                             'different, got {} and {}.'.format(n_stats, n_conts))

    n_rows = n_conts
    n_cols = 3 # because I am doing the ortho plot

    figsize = figsize[0] * n_cols, figsize[1] * n_rows
    fig = plt.figure(figsize=figsize, facecolor='black')
    gs  = gridspec.GridSpec(n_rows, 1)

    # put the title, if any
    if title:
        fig.suptitle(title, fontsize=title_fontsize, color='white')

    for i, (simg, cimg) in enumerate(zip(_stat_imgs, _cnts_imgs)):
        ax = plt.subplot(gs[i])
        p = plot_stat_overlay(simg, cimg, bg_img=bg_img, figure=fig, axes=ax, **kwargs)

    return fig



from .image import (concat_imgs,
                    math_img,
                    mean_img,
                    smooth_img,
                    copy_header,
                    resample,
                    resample_to_img,
                    ni2file)

from .canica import CanICAInterface

from .connectivity import ConnectivityCorrelationInterface

from .plot import (plot_all_components,
                   plot_ica_components,
                   plot_multi_slices,
                   plot_ortho_slices)

from .roi import spread_labels


# -*- coding: utf-8 -*-
"""
Function utilities to deal with spatial/coordinates problems in the nifti images.
"""


def get_bounding_box(in_file):
    """ Retrieve the bounding box of a volume in millimetres."""

    # the imports must be inside if you want them to work in a nipype.Function node.
    from itertools import product
    import nibabel as nib
    import numpy   as np

    img = nib.load(in_file)

    # eight corners of the 3-D unit cube [0, 0, 0] .. [1, 1, 1]
    corners = np.array(list(product([0, 1], repeat=3)))
    # scale to the index range of the volume
    corners = corners * (np.array(img.shape[:3]) - 1)
    # apply the affine transform
    corners = img.affine.dot(np.hstack([corners, np.ones((8, 1))]).T).T[:, :3]

    # get the extents
    low_corner  = np.min(corners, axis=0)
    high_corner = np.max(corners, axis=0)

    return [low_corner.tolist(), high_corner.tolist()]


# -*- coding: utf-8 -*-
"""
PET-only image registration nipype workflow.
"""
import os.path as op

from   ..preproc import spm_warp_to_mni
from   .._utils  import format_pair_list
from   ..utils   import (get_datasink,
                         extend_trait_list,
                         get_input_node,
                         remove_ext,
                         get_input_file_name,
                         extension_duplicates)


def attach_spm_pet_preprocessing(main_wf, wf_name='spm_pet_preproc'):
    """ Attach a FDG-PET only pre-processing workflow that uses SPM12 to `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.select.pet: input node

    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    # The base name of the 'pet' file for the substitutions
    pet_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'pet')))

    # get the PET preprocessing pipeline
    warp_pet_wf = spm_warp_to_mni(wf_name=wf_name)

    # dataSink output substitutions
    regexp_subst = [
                    (r'/w{pet}.nii', '/{pet}_mni.nii'),
                   ]
    regexp_subst = format_pair_list(regexp_subst, pet=pet_fbasename)
    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    # Connect the nodes
    main_wf.connect([
                     # pet file input
                    (in_files,    warp_pet_wf, [('pet',                      'warp_input.in_files')]),
                    (warp_pet_wf, datasink,    [('warp_output.warped_files', 'pet.@warped'),]),
                   ])

    return main_wf



# -*- coding: utf-8 -*-
"""
Helper functions to plot connectivity results.
"""


def plot_connectivity_matrix(connectivity_matrix, label_names):
    """ Plot the connectivity matrix. """
    # Display the correlation matrix
    import numpy as np
    from matplotlib import pyplot as plt

    plt.figure(figsize=(10, 10))

    # Mask out the major diagonal
    np.fill_diagonal(connectivity_matrix, 0)
    fig = plt.imshow(connectivity_matrix, interpolation="nearest", cmap="RdBu_r",
                     vmax=0.8, vmin=-0.8)
    plt.colorbar()
    # And display the labels
    _ = plt.xticks(range(len(label_names)), label_names, rotation=90)
    _ = plt.yticks(range(len(label_names)), label_names)

    return fig



"""
Function utilities that use pandas.
"""

import pandas as pd


def add_table_headers(values, axis0_labels, axis1_labels, **kwargs):
    """ Return a pandas.DataFrame with the content of `values`.
    This DataFrame will have each row labeled by `axis0_labels`, and
    each column by `axis1_labels`.

    Parameters
    ----------
    values: np.ndarray
        MxN matrix
        M is the length of axis0_labels.
        N is the lenght of axis1_labels.

    axis0_labels: array or any type
        This will be used as index in the DataFrame construction.

    axis1_labels: array or any type
        This will be used as column in the DataFrame construction.

    kwargs: keyword arguments
        Additional columns to be added to the resulting DataFrame.
        The argument names are the column name and the values must be sequences of length M.

    Returns
    -------
    labeled_measure: pandas.DataFrame

    Notes
    -----
    My previous solution was more complex and I had this function
    prepared before realizing this was ridiculously simple.
    """
    df = pd.DataFrame(values, index=axis0_labels, columns=axis1_labels)

    for k, v in kwargs.items():
        if len(v) != len(axis0_labels):
            raise AttributeError('The value for argument {} should have length {} but has '
                                 'length {}.'.format(k, len(axis0_labels), len(v)))
        df[k] = v

    return df


def write_tabbed_excel(filepath, dict_df, **kwargs):
    """ Write each df in `dict_df` in a tab of the Excel spreadsheet in `filepath`.

    Parameters
    ----------
    filepath: str

    dict_df: Dict[pandas.DataFrame]

    kwargs:
        columns: list of str
            List of the columns of `df` to be written in the sheet.

        other arguments to be passed to the df.to_excel function.
    """
    from pandas import ExcelWriter

    with ExcelWriter(filepath) as writer:
        for key, df in dict_df.items():
            # write df to the tab
            df.to_excel(writer, sheet_name=key, **kwargs)


# -*- coding: utf-8 -*-
"""
Utilities to use nilearn.image from nipype
"""


def ni2file(**presuffixes):
    from functools import wraps

    """ Pick the nibabel image container output from `f` and stores it in a file.
    If the shape attribute of the container is True, will save it into a file, otherwise
    will directly return the scalar value.

    To know the path where it has to save the file it looks into argument
    values in this order:
    - `out_file` kwarg in the call to `f`,
    - `out_file` kwarg in the `ni2file` decorator,
    - the first argument in the function call.
    In the last case a presuffix must be defined in the decorator to avoid overwriting
    an existing file.
    """
    def _pick_an_input_file(*args, **kwargs):
        """Assume that either the first arg or the first kwarg is an input file."""
        if args:
            return args[0]
        else:
            return list(kwargs.values())[0]

    def nifti_out(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            res_img = f(*args, **kwargs)
            if isinstance(res_img, list):
                if len(res_img) == 1:
                    res_img = res_img[0]
                else:
                    return res_img

            if not res_img.shape: # the result is a scalar value
                return res_img.get_data().flatten()[0]

            import os.path as op
            from nipype.utils.filemanip import fname_presuffix

            out_file = kwargs.get('out_file', presuffixes.pop('out_file', None))
            if out_file is not None:
                if not presuffixes and op.exists(out_file):
                    raise IOError('The file {} already exists, please add a presuffix to the'
                                  'decorator.'.format(out_file))

                out_file = fname_presuffix(out_file, **presuffixes)
            else:
                in_file = kwargs.get('in_file', None)
                if in_file is None:
                    in_file = _pick_an_input_file(*args, **kwargs)

                if not op.exists(in_file):
                    raise IOError('Expected an existing file to use as reference for'
                                  ' the output file name, got {}.'.format(in_file))

                out_file = fname_presuffix(op.basename(in_file), **presuffixes)

            if not out_file:
                raise ValueError("Could not find a output file name for this function: "
                                " {}({}, {}).".format(f.__name__, *args, **kwargs))

            res_img.to_filename(out_file)

            return op.abspath(out_file)

        return wrapped
    return nifti_out


@ni2file(out_file='nilearn_maths.nii.gz')
def math_img(formula, out_file='', **imgs):
    """ Use nilearn.image.math_img.
    This function in addition allows imgs to contain numerical scalar values.

    Returns
    -------
    out_file: str
        The absolute path to the output file.
    """
    import numpy as np
    import nilearn.image as niimg
    from   six import string_types

    for arg in list(imgs.keys()):
        if isinstance(imgs[arg], string_types):
            continue

        if np.isscalar(imgs[arg]):
            if arg not in formula:
                raise ValueError("Could not find {} in the formula: {}.".format(arg, formula))

            formula = formula.replace(arg, str(imgs[arg]))
            imgs.pop(arg)

    return niimg.math_img(formula=formula, **imgs)


@ni2file(suffix='_resampled')
def resample(in_file, **kwargs):
    """ Use nilearn.image.resample_img.

    Returns
    -------
    out_file: str
        The absolute path to the output file.
    """
    import nilearn.image as niimg
    return niimg.resample_img(img=in_file, **kwargs)


@ni2file(suffix='_resampled')
def resample_to_img(in_file, target, **kwargs):
    """ Use nilearn.image.resample_to_img.

    Returns
    -------
    out_file: str
        The absolute path to the output file.
    """
    import nilearn.image as niimg
    return niimg.resample_to_img(source_img=in_file, target_img=target, **kwargs)


@ni2file(out_file='concat_img.nii.gz')
def concat_imgs(in_files, out_file=None):
    """ Use nilearn.image.concat_imgs to concat images of up to 4 dimensions.
    Returns
    -------
    out_file: str
        The absolute path to the output file.
    """
    import nilearn.image as niimg
    return niimg.concat_imgs(in_files)


@ni2file(out_file='concat_img.nii.gz')
def concat_3D_imgs(in_files, out_file=None):
    """ Use nilearn.image.concat_imgs to concat 3D volumes into one 4D volume.

    If `in_files` is a list of 3D volumes the return value is the path to one 4D volume.
    Else if `in_files` is a list of 4D volumes the return value is `in_files`.

    Returns
    -------
    out_file: str
        The absolute path to the output file.
    """
    import nilearn.image as niimg

    from   nilearn._utils import check_niimg_3d

    all_3D = True
    for idx, img in enumerate(in_files):
        try:
            _ = check_niimg_3d(img)
        except Exception:
            all_3D = False
            break

    if not all_3D:
        #raise AttributeError('Expected all input images to be 3D volumes, but '
        #                     ' at least the {}th is not.'.format(idx))
        return in_files
    else:
        return niimg.concat_imgs(in_files)


@ni2file(suffix='_mean')
def mean_img(in_file, out_file=None):
    """ Use nilearn.image.mean_img.
    Returns
    -------
    out_file: str
        The absolute path to the output file.
    """
    import nilearn.image as niimg
    return niimg.mean_img(in_file)


@ni2file(suffix='_smooth')
def smooth_img(in_file, fwhm, out_file=None):
    """ Use nilearn.image.smooth_img.
    Returns
    -------
    out_file: str
        The absolute path to the output file.
    """
    import nilearn.image as niimg
    return niimg.smooth_img(in_file, fwhm=fwhm)


@ni2file(suffix='')
def copy_header(in_file, data_file, out_file=None):
    """ Use nilearn.image.new_img_like to copy the header
    from `in_file` to `data_file` and return the result.
    Returns
    -------
    out_file: str
        The absolute path to the output file.
    """
    import nilearn.image as niimg

    img = niimg.load_img(data_file)

    return niimg.new_img_like(in_file, img.get_data(),
                              affine=img.affine, copy_header=True)


# -*- coding: utf-8 -*-
"""
Nipype workflows to co-register anatomical MRI to diffusion MRI.
"""
import nipype.pipeline.engine as pe
from   nipype.interfaces.fsl import MultiImageMaths
from   nipype.interfaces.utility import IdentityInterface, Select, Split
from   nipype.algorithms.misc import Gunzip

from .._utils  import flatten_list
from ..config  import setup_node, check_atlas_file
from ..preproc import spm_coregister


def spm_anat_to_diff_coregistration(wf_name="spm_anat_to_diff_coregistration"):
    """ Co-register the anatomical image and other images in anatomical space to
    the average B0 image.

    This estimates an affine transform from anat to diff space, applies it to
    the brain mask and an atlas.

    Nipype Inputs
    -------------
    dti_co_input.avg_b0: traits.File
        path to the average B0 image from the diffusion MRI.
        This image should come from a motion and Eddy currents
        corrected diffusion image.

    dti_co_input.anat: traits.File
        path to the high-contrast anatomical image.

    dti_co_input.tissues: traits.File
        paths to the NewSegment c*.nii output files, in anatomical space

    dti_co_input.atlas_anat: traits.File
        Atlas in subject anatomical space.

    Nipype Outputs
    --------------
    dti_co_output.anat_diff: traits.File
        Anatomical image in diffusion space.

    dti_co_output.tissues_diff: traits.File
        Tissues images in diffusion space.

    dti_co_output.brain_mask_diff: traits.File
        Brain mask for diffusion image.

    dti_co_output.atlas_diff: traits.File
        Atlas image warped to diffusion space.
        If the `atlas_file` option is an existing file and `normalize_atlas` is True.

    Nipype Workflow Dependencies
    ----------------------------
    This workflow depends on:
    - spm_anat_preproc

    Returns
    -------
    wf: nipype Workflow
    """
    # specify input and output fields
    in_fields  = ["avg_b0", "tissues", "anat"]
    out_fields = ["anat_diff",
                  "tissues_diff",
                  "brain_mask_diff",]

    do_atlas, _ = check_atlas_file()
    if do_atlas:
        in_fields  += ["atlas_anat"]
        out_fields += ["atlas_diff"]

    # input interface
    dti_input = pe.Node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                        name="dti_co_input")

    gunzip_b0 = pe.Node(Gunzip(), name="gunzip_b0")
    coreg_b0  = setup_node(spm_coregister(cost_function="mi"), name="coreg_b0")

    # co-registration
    brain_sel    = pe.Node(Select(index=[0, 1, 2]),            name="brain_sel")
    coreg_split  = pe.Node(Split(splits=[1, 2], squeeze=True), name="coreg_split")

    brain_merge  = setup_node(MultiImageMaths(), name="brain_merge")
    brain_merge.inputs.op_string = "-add '%s' -add '%s' -abs -kernel gauss 4 -dilM -ero -kernel gauss 1 -dilM -bin"
    brain_merge.inputs.out_file = "brain_mask_diff.nii.gz"

    # output interface
    dti_output = pe.Node(IdentityInterface(fields=out_fields),
                         name="dti_co_output")

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # Connect the nodes
    wf.connect([
                # co-registration
                (dti_input, coreg_b0, [("anat", "source")]),

                (dti_input,     brain_sel,   [("tissues",             "inlist")]),
                (brain_sel,     coreg_b0,    [(("out", flatten_list), "apply_to_files")]),

                (dti_input,     gunzip_b0,   [("avg_b0",   "in_file")]),
                (gunzip_b0,     coreg_b0,    [("out_file", "target")]),

                (coreg_b0,      coreg_split, [("coregistered_files", "inlist")]),
                (coreg_split,   brain_merge, [("out1",               "in_file")]),
                (coreg_split,   brain_merge, [("out2",               "operand_files")]),

                # output
                (coreg_b0,     dti_output,     [("coregistered_source", "anat_diff")]),
                (coreg_b0,     dti_output,     [("coregistered_files",  "tissues_diff")]),
                (brain_merge,  dti_output,     [("out_file",            "brain_mask_diff")]),
              ])

    # add more nodes if to perform atlas registration
    if do_atlas:
        coreg_atlas = setup_node(spm_coregister(cost_function="mi"), name="coreg_atlas")

        # set the registration interpolation to nearest neighbour.
        coreg_atlas.inputs.write_interp = 0
        wf.connect([
                    (dti_input,   coreg_atlas, [("anat",               "source"),
                                                ("atlas_anat",         "apply_to_files"),
                                               ]),
                    (gunzip_b0,   coreg_atlas, [("out_file",           "target")]),
                    (coreg_atlas, dti_output,  [("coregistered_files", "atlas_diff")]),
                  ])

    return wf




# -*- coding: utf-8 -*-
"""
Resting state networks atlas container class
"""
from collections import OrderedDict

import nilearn.plotting as niplot
import nilearn.image as niimg


class RestingStateNetworks:
    """ A container class to parse and return useful values/images
    from RSN templates.

    Parameters
    ----------
    img_file: str
        Path to a img-like RSN templates file.

    txt_file: str
        Path to a text file with the labels for `img_file`.
        The expected structure of each line of
        this file is  <name of the RSN>, <list of indices>.
        For example:

        Basal Ganglia, 21
        Auditory, 17
        Sensorimotor, 7, 23, 24, 38, 56, 29
        Visual, 46, 64, 67, 48, 39, 59
        Default-Mode, 50, 53, 25, 68
        Attentional, 34, 60, 52, 72, 71, 55
        Frontal, 42, 20, 47, 49

    start_from_one: bool
        If True it means that the `txt_file` volume indices start from 1.
        From 0 otherwise. Be careful, the default is True!
    """
    def __init__(self, img_file, txt_file, start_from_one=True):
        self._img_file       = img_file
        self._txt_file       = txt_file
        self._start_from_one = start_from_one

        self.network_names = self._network_names()
        self._img          = niimg.load_img(self._img_file)
        self._self_check()

    def iter_networks(self):
        """Yield idx (what is in the text_file) and
        image of each RSN volume from the text file."""
        for idx, name in self.network_names.items():
            yield idx, self._get_img(idx)

    def iter_networks_names(self):
        """Yield idx (what is in the text_file) and
        image of each RSN volume from the text file."""
        for idx, name in self.network_names.items():
            yield idx, name

    def _network_names(self):
        """Return OrderedDict[int]->str, with the index and the name of each
        RSN."""
        names_idx = self._read_labels_file()
        return OrderedDict([(idx, name) for name, idxs in names_idx.items()
                                        for idx in idxs])

    def _get_img(self, network_index):
        """ Return one RSN given the index in the labels file."""
        img_idx = self._img_index(network_index)
        return niimg.index_img(self._img, img_idx)

    def _img_index(self, network_index):
        """Return the correspoding image index for the given network index."""
        if self._start_from_one:
            return network_index - 1

        return network_index

    def _self_check(self):
        """Simple content check."""
        n_labels = len(self.network_names)
        n_images = self._img.shape[-1]

        if n_labels == n_images:
            return

        if n_labels > n_images:
            raise ValueError('The number of labels is larger than the number '
                             'of images. Got {} and {}.'.format(n_labels, n_images))

        # print('The number of volumes in the image is different from the number '
        #       'of labels in the text file.\n I am going to use only the ones '
        #       ' in the text file.')

    def _read_labels_file(self):
        """ Read the text file and return a dict[str->List[int]] with network
        names and blob indices.
        """
        lines = [l.rstrip('\n') for l in open(self._txt_file).readlines()]

        netblobs = OrderedDict()
        for l in lines:
            pcs = l.split(',')
            netname = pcs[0]
            blobs   = [int(idx) for idx in pcs[1:]]

            netblobs[netname] = blobs
        return netblobs

    def plot_all(self):
        names = self.network_names
        for idx, rsn in enumerate(niimg.iter_img(self._img)):
            disp = niplot.plot_roi(rsn, title=names.get(idx, None))

    def join_networks(self, network_indices):
        """Return a NiftiImage containing a binarised version of the sum of
        the RSN images of each of the `network_indices`."""
        oimg = self._get_img(network_indices[0]).get_data()
        for idx in network_indices[1:]:
            oimg += self._get_img(idx).get_data()

        return niimg.new_img_like(self._get_img(network_indices[0]),
                                  oimg.astype(bool))

    def __iter__(self):
        return (img for idx, img in self.iter_networks())

    def __len__(self):
        return len(self.network_names)




# -*- coding: utf-8 -*-
"""
Private helper functions
"""


def grep(lines, substr):
    """ Return a list of strings from `lines` that have
    `substr` as a substring.
    """
    return [l for l in lines if substr in l]


def check_equal(lst):
    """ Return True if all items in `lst` are equal, False otherwise.
    Note that check_equal([1, True]) is True.
    """
    return not lst or lst.count(lst[0]) == len(lst)


def flatten_list(list_of_lists):
    """ Will convert a list of lists in a list with the items inside each sub-list.

    Parameters
    ----------
    list_of_lists: list[list[object]]

    Returns
    -------
    list
    """
    if not list_of_lists:
        return []

    if isinstance(list_of_lists[0], list):
        lst = []
        for l in list_of_lists:
            lst.extend(l)
        return lst
    return list_of_lists


def format_pair_list(pair_list, **kwargs):
    """ Given a list of 2-tuples of str, calls format with `kwargs` for each
    item in the 2-tuples and return the formatted list of 2-tuples.

    Parameters
    ----------
    pair_list: list of 2-tuples of str

    kwargs: keyword arguments
        Arguments for the format function of each string in the 2-tuples.

    Returns
    -------
    formatted_pair_list: list of 2-tuples of str
    """
    return [(s[0].format(**kwargs), s[1].format(**kwargs)) for s in pair_list]


def _check_list(str_or_list):
    """ If `str_or_list` is a list will return it as it is. If it is a str, will return a
    list with the str inside.

    Parameters
    ----------
    str_or_list: str or list

    Returns
    -------
    list
    """
    if str_or_list is None:
        return None

    if isinstance(str_or_list, list):
        return str_or_list

    if isinstance(str_or_list, str):
        return [str_or_list]

    raise ValueError('Expected a `str` or a `list`, ' \
                     'got {}.'.format(type(str_or_list)))


def get_values_map_keys(records, keyidx=0):
    """ Given a dict of str->2-tuples, e.g.:
            {'anat': [('modality', 'anat'), ('image_file', 'anat_hc.nii.gz')],
             'pet':  [('modality', 'pet'),  ('image_file', 'pet_fdg.nii.gz')],

        or

       Given a list of list of 2-tuples of str, e.g.:
            [[('modality', 'anat'), ('image_file', 'anat_hc.nii.gz')],
              ('modality', 'pet'),  ('image_file', 'pet_fdg.nii.gz')],


    Will return the unique values of each record value, in this case:
    {'modality', 'image_file'}.

    Parameters
    ----------
    values_maps_dict: Dict[str->2-tuple]

    Returns
    -------
    keys: set[str]
    """
    if not records or records is None:
        return []

    if isinstance(records, dict):
        itemset = records.values()
    elif isinstance(records, list):
        itemset = records
    else:
        raise NotImplementedError('Expected a `dict` or a `list of list` as `records, '
                                  'got {}.'.format(type(records)))

    crumb_args = set()
    for items in itemset:
        crumb_args = crumb_args.union(set([t[keyidx] for t in items]))

    return crumb_args


# -*- coding: utf-8 -*-
"""
Functions to help comparing and dealing with spatial maps.
"""
import itertools

import numpy as np
import nilearn.masking as nimask
import nilearn.image as niimg
from   sklearn.metrics.pairwise import pairwise_distances
from   boyle.nifti.utils import nifti_out, thr_img, icc_img_to_zscore


@nifti_out
def spatial_map(icc, thr, mode='+'):
    """ Return the thresholded z-scored `icc`. """
    return thr_img(icc_img_to_zscore(icc), thr=thr, mode=mode).get_data()


def spatial_maps_pairwise_similarity(imgs1, imgs2, mask_file, distance='correlation'):
    """ Similarity values of each image in `imgs1` to each image in `imgs2`, both masked by `mask_file`.
    These values are based on distance metrics, specified by `distance` argument.
    The resulting similarity value is the complementary value of the distance,
    i.e., '1 - <distance value>'.
    The images in `imgs1` will be resampled to `imgs2` if their affine matrix don't match.

    Parameters
    ----------
    imgs1: list of niimg-like or 4D niimg-like

    imgs2: list of niimg-like or 4D niimg-like

    mask_file: niimg-like

    distance: str
        Valid values for `distance` are:
        From scikit-learn: ['cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan'].
        From scipy.spatial.distance: ['braycurtis', 'canberra', 'chebyshev', 'correlation', 'dice', 'hamming',
                                      'jaccard', 'kulsinski', 'mahalanobis', 'matching', 'minkowski',
                                      'rogerstanimoto', 'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath',
                                      'sqeuclidean', 'yule']
                                      See the documentation for scipy.spatial.distance for details on these metrics.

    Returns
    -------
    corrs: np.ndarray
        A matrix of shape MxN, where M is len(imgs1) and N is len(imgs2).
        It contains the similarity values.
    """
    img1_ = niimg.load_img(imgs1)
    img2_ = niimg.load_img(imgs2)

    n_imgs1 = img1_.shape[-1]
    n_imgs2 = img2_.shape[-1]
    corrs = np.zeros((n_imgs1, n_imgs2), dtype=float)

    mask_trnsf = niimg.resample_to_img(mask_file, niimg.index_img(img2_, 0),
                                       interpolation='nearest',
                                       copy=True)

    for idx1, img1 in enumerate(niimg.iter_img(img1_)):
        img1_resamp = niimg.resample_to_img(img1, niimg.index_img(img2_, 0), copy=True)
        img1_masked = nimask.apply_mask(img1_resamp, mask_trnsf)

        for idx2, img2 in enumerate(niimg.iter_img(img2_)):
            img2_masked = nimask.apply_mask(img2, mask_trnsf)
            dist = pairwise_distances(img1_masked.reshape(1, -1),
                                      img2_masked.reshape(1, -1),
                                      metric=distance)

            # since this is a scalar value
            dist = dist[0][0]

            # since this is a distance, not a similarity value
            corr = 1 - dist

            # store it
            corrs[idx1, idx2] = corr

    return corrs


def spatial_maps_goodness_of_fit(rsn_imgs, spatial_maps, mask_file, rsn_thr=4.0):
    """ Goodness-of-fit values described as in Zhou et al., 2010, Brain.

    Parameters
    ----------
    rsn_imgs: list of niimg-like or 4D niimg-like
        The RSN maps. They should be thresholded beforehand if `rsn_thr` is lower or equal than 0.

    spatial_maps: list of niimg-like or 4D niimg-like

    mask_file: niimg-like
        An extra mask to apply to the thresholded RSN masks.
        This is used to exclude values outside of the RSN blobs.
        It is recommended to use a brain mask for this.

    rsn_thr: float, optional
        The threshold to apply to `rsn_imgs` to create the RSN masks.
        If rsn_thr <= 0, no thresholding will be applied.

    Returns
    -------
    gof_df: np.ndarray
        A matrix of shape MxN, where M is len(rsn_imgs) and N is len(spatial_maps).
        It contains the goodness-of-fit values.

    Notes
    -----
    "These ICN templates were thresholded at a z-score 4.0 to be visually comparable to the
    consistent ICNs published by Damoiseaux et al. (2006). A minor modification of previous
    goodness-of-fit methods (Seeley et al., 2007b, 2009) was included here for template
    matching, with goodness-of-fit scores calculated by multiplying
    (i) the average z-score difference between voxels falling within the template and
    voxels falling outside the template; and
    (ii) the difference in the percentage of positive z-score voxels inside and outside the template.

    This goodness-of-fit algorithm proves less vulnerable to inter-subject variability in shape,
    size, location and strength of each ICN", than the one published in Seeley et al., 2007b.
    This latter method only uses the value in (i).

    Extracted from Zhou et al., 2010, Brain.
    """
    rsn_img = niimg.load_img(rsn_imgs)
    spm_img = niimg.load_img(spatial_maps)

    n_rsns = rsn_img.shape[-1]
    n_ics  =  spm_img.shape[-1]
    gofs   = np.zeros((n_rsns, n_ics), dtype=float)

    # threshold the RSN templates
    if rsn_thr > 0:
        thr_rsns = (spatial_map(rsn, thr=rsn_thr, mode='+-')
                    for rsn in niimg.iter_img(rsn_img))
    else:
        thr_rsns = rsn_img

    # for each RSN template and IC image
    iter_rsn_ic = itertools.product(enumerate(niimg.iter_img(thr_rsns)),
                                    enumerate(niimg.iter_img( spm_img)))

    for (rsn_idx, rsn), (ic_idx, ic) in iter_rsn_ic:

        ref_vol = rsn.get_data()
        rsn_vol = np.zeros(rsn.shape, dtype=int)

        #rsn_in  = niimg.math_img('np.abs(img) > 0', img=rsn)
        rsn_in = rsn_vol.copy()
        rsn_in[np.abs(ref_vol) > 0] = 1

        #rsn_out = niimg.math_img('img == 0', img=rsn)
        rsn_out = rsn_vol.copy()
        rsn_out[ref_vol == 0] = 1

        if mask_file is not None:
            # rsn_out = niimg.math_img('mask * img', mask=rsn_brain_mask, img=rsn_out)
            rsn_brain_mask = niimg.resample_to_img(mask_file, rsn,
                                                   interpolation='nearest')
            rsn_out = rsn_brain_mask.get_data() * rsn_out

        # convert the mask arrays to image in order to resample
        rsn_in  = niimg.new_img_like(rsn, rsn_in)
        rsn_out = niimg.new_img_like(rsn, rsn_out)

        rsn_in  = niimg.resample_to_img(rsn_in,  ic, interpolation='nearest')
        rsn_out = niimg.resample_to_img(rsn_out, ic, interpolation='nearest')

        # apply the mask
        #zscore_in  = niimg.math_img('mask * img', mask=rsn_in,  img=ic).get_data()
        zscore_in = rsn_in.get_data() * ic.get_data()

        #zscore_out = niimg.math_img('mask * img', mask=rsn_out, img=ic).get_data()
        zscore_out = rsn_out.get_data() * ic.get_data()

        #gof_term1
        # calculate the the average z-score difference between voxels falling
        # within the template and voxels falling outside the template
        gof_term1 = zscore_in.mean() - zscore_out.mean()

        #gof_term2
        # the difference in the percentage of positive z-score voxels inside and outside the template.
        n_pos_zscore_in  = np.sum(zscore_in  > 0)
        n_pos_zscore_out = np.sum(zscore_out > 0)
        n_pos_zscore_tot = n_pos_zscore_in + n_pos_zscore_out

        if n_pos_zscore_tot != 0:
            n_pos_zscore_pcnt = 100 / n_pos_zscore_tot
            gof_term2 = (n_pos_zscore_in - n_pos_zscore_out) * n_pos_zscore_pcnt
        else:
            gof_term2 = 0

        # global gof
        gof = gof_term1 * gof_term2

        # add the result
        gofs[rsn_idx][ic_idx] = gof

    return gofs


def nd_vector_correlations(data, vector, n=4):
    """ Calculate the pairwise correlation between `vector` and each element
    of the last dimension of `data`.

    Parameters
    ----------
    data: numpy.ndarray
        nD array

    vector: numpy.array
        1D vector

    n: int

    Returns
    -------
    corrs: numpy.array
        (n-1)D array
    """
    assert data.ndim == n
    assert data.shape[-1] == list(vector.shape)[0]

    trid_shape = data.shape[:n-1]

    n_voxels = np.prod(trid_shape)
    n_ts     = data.shape[-1]

    fourthd_vecs = data.reshape(n_voxels, n_ts)
    corrs        = - pairwise_distances(fourthd_vecs, vector[np.newaxis], 'correlation') + 1
    corrs[np.isnan(corrs)] = 0

    return corrs.reshape(trid_shape)


from sklearn.linear_model import Ridge


def _compute_loadings(components, data):
    """
    DONT USE THIS YET!

    Parameters
    ----------
    components:

    data:

    Returns
    -------

    """
    ridge = Ridge(fit_intercept=None, alpha=1e-8)
    ridge.fit(components.T, np.asarray(data.T))
    loadings = ridge.coef_.T

    # unit-variance scaling
    S = np.sqrt(np.sum(loadings ** 2, axis=0))
    S[S == 0] = 1
    loadings /= S[np.newaxis, :]
    return loadings


def spatiotemporal_regression(components, imgs, mask, confounds=None):
    """Project the data into a reduced representation

    DONT USE THIS YET!

    Parameters
    ----------
    components:

    imgs: iterable of Niimg-like objects
        See http://nilearn.github.io/manipulating_images/input_output.html.
        Data to be projected

    confounds: CSV file path or 2D matrix
        This parameter is passed to nilearn.signal.clean. Please see the
        related documentation for details

    Returns
    ----------
    loadings: list of 2D ndarray,
        For each subject, each sample, loadings for each decomposition
        components
        shape: number of subjects * (number of scans, number of regions)

    """
    components_img_ = self.masker_.inverse_transform(self.components_)

    nifti_maps_masker = NiftiMapsMasker(
        components_img_, self.masker_.mask_img_,
        resampling_target='maps')

    nifti_maps_masker.fit()

    if confounds is None:
        confounds = [None] * len(imgs)

    return [nifti_maps_masker.transform(img, confounds=confound)
            for img, confound in zip(imgs, confounds)]


def inverse_transform(self, loadings):
    """Use provided loadings to compute corresponding linear component
    combination in whole-brain voxel space

    DONT USE THIS YET

    Parameters
    ----------
    loadings: list of numpy array (n_samples x n_components)
        Component signals to tranform back into voxel signals

    Returns
    -------
    reconstructed_imgs: list of nibabel.Nifti1Image
        For each loading, reconstructed Nifti1Image

    """
    if not hasattr(self, 'components_'):
        ValueError('Object has no components_ attribute. This is either '
                   'because fit has not been called or because'
                   '_DecompositionEstimator has directly been used')
    self._check_components_()
    components_img_ = self.masker_.inverse_transform(self.components_)
    nifti_maps_masker = NiftiMapsMasker(
        components_img_, self.masker_.mask_img_,
        resampling_target='maps')
    nifti_maps_masker.fit()
    # XXX: dealing properly with 2D/ list of 2D data?
    return [nifti_maps_masker.inverse_transform(loading)
            for loading in loadings]




# -*- coding: utf-8 -*-
"""
PET image preprocessing utilities nipype function helpers.
"""
from   nipype.interfaces.base import traits
from   nipype.interfaces.utility import Merge, Function, IdentityInterface
from   nipype.pipeline import Workflow

from   ..interfaces.nilearn import math_img, concat_imgs, resample_to_img
from   ..config  import setup_node
from   ..preproc import PETPVC
from   ..utils   import selectindex, rename


#TODO: add a becquerel/ml normalization function node
# hint: img*slope + intercept


def petpvc_cmd(in_file=traits.Undefined, mask_file=traits.Undefined, out_file=traits.Undefined,
               pvc_method='RBV', fwhm_mm=(4, 4, 4)):
    """ Return a nipype interface to PETPVC.

    Parameters
    ----------
    in_file: str
        Path to the PET image file.

    out_file: str
        Path to the result output file

    mask_file: str
        Path to the mask image file with tissue maps.

    pvc_method: str
        Keyword of the method to run.
        Run `petpvc --help` for choices.

    fwhm_mm: iterable of floats
        Iterable with 3 ints or floats to define the full-width at half maximum in mm along
        each axis of the point-spread function (PSF) of the scanner.

    Returns
    -------
    pet_pvc: PETPVC(nipype.interfaces.base.CommandLine)
    """

    pvc = PETPVC()
    pvc.inputs.in_file   = in_file
    pvc.inputs.out_file  = out_file
    pvc.inputs.mask_file = mask_file
    pvc.inputs.pvc       = pvc_method
    pvc.inputs.fwhm_x    = fwhm_mm[0]
    pvc.inputs.fwhm_y    = fwhm_mm[1]
    pvc.inputs.fwhm_z    = fwhm_mm[2]

    return pvc


def petpvc_mask(wf_name="petpvc_mask"):
    """ A Workflow that returns a 4D merge of 4 volumes for PETPVC: GM, WM, CSF and background.

    Parameters
    ----------
    wf_name: str
        The name of the workflow.

    Nipype.Inputs
    -------------
    pvcmask_input.tissues: list of existing files
        List of tissue files in anatomical space, the 3 file
        paths must be in this order: GM, WM, CSF

    Nipype.Outputs
    --------------
    pvcmask_output.petpvc_mask: existing file
        A 4D volume file with these maps in order: GM, WM, CSF, background

    pvcmask_output.brain_mask: existing file
        A mask that is a binarised sum of the tissues file with fslmaths.
        Can be used as brain mask in anatomical space for the PET image.

    Returns
    -------
    wf: nipype Workflow
    """
    # define nodes
    # specify input and output fields
    in_fields  = ["tissues"]

    out_fields = ["petpvc_mask",
                  "brain_mask",]

    # input
    pvcmask_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                               name="pvcmask_input")

    tissues = setup_node(IdentityInterface(fields=["gm", "wm", "csf"], mandatory_inputs=True),
                         name="tissues")

    merge_list = setup_node(Merge(4), name="merge_list")

    ## maths for background
    img_bkg = setup_node(Function(function=math_img,
                                  input_names=["formula", "out_file", "gm", "wm", "csf"],
                                  output_names=["out_file"],
                                  imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                          name='background')
    img_bkg.inputs.out_file = "tissue_bkg.nii.gz"
    img_bkg.inputs.formula  = "np.maximum((-((gm + wm + csf) - 1)), 0)"

    ## maths for brain mask
    brain_mask = setup_node(Function(function=math_img,
                                     input_names=["formula", "out_file", "gm", "wm", "csf"],
                                     output_names=["out_file"],
                                     imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                            name='brain_mask')
    brain_mask.inputs.out_file = "tissues_brain_mask.nii.gz"
    brain_mask.inputs.formula  = "np.abs(gm + wm + csf) > 0"

    ## concat the tissues images and the background for PETPVC
    merge_tissues = setup_node(Function(function=concat_imgs,
                                        input_names=["in_files"],
                                        output_names=["out_file"],
                                        imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                               name='merge_tissues')
    merge_tissues.inputs.out_file = "petpvc_mask.nii.gz"

    # output
    pvcmask_output = setup_node(IdentityInterface(fields=out_fields), name="pvcmask_output")

    # Create the workflow object
    wf = Workflow(name=wf_name)

    # Connect the nodes
    wf.connect([
                # separate [GM, WM, CSF] into [GM] and [WM, CSF]
                (pvcmask_input, tissues, [(("tissues", selectindex, 0), "gm"),
                                          (("tissues", selectindex, 1), "wm"),
                                          (("tissues", selectindex, 2), "csf"),
                                         ]),

                (tissues, img_bkg,    [("gm", "gm" ), ("wm", "wm" ), ("csf", "csf"),]),
                (tissues, brain_mask, [("gm", "gm" ), ("wm", "wm" ), ("csf", "csf"),]),
                (tissues, merge_list, [("gm", "in1"), ("wm", "in2"), ("csf", "in3"),]),

                # create a list of [GM, WM, CSF, BKG]
                (img_bkg, merge_list, [("out_file", "in4")]),

                # merge into 4D: [GM, WM, CSF, BKG]
                (merge_list, merge_tissues, [("out", "in_files")]),

                # output
                (merge_tissues, pvcmask_output, [("out_file", "petpvc_mask")]),
                (brain_mask,    pvcmask_output, [("out_file", "brain_mask")]),
              ])

    return wf


def intensity_norm(wf_name='intensity_norm'):
    """ Workflow that uses a mask against a source from where the mean value will be taken.
    This mean value will be used to demean the whole source and leave it in out_file.

    Parameters
    ----------
    wf_name: str
        The name of the workflow.

    Nipype Inputs
    -------------
    intnorm_input.source: existing file
        The image from where to extract the signal values and normalize.

    intnorm_input.mask: existing file
        The mask to specify which voxels to use to calculate the statistics
        for normalization.

    Nipype Outputs
    --------------
    intnorm_output.out_file: existing file

    Returns
    -------
    wf: nipype Workflow
    """
    # specify input and output fields
    in_fields  = ["source",
                  "mask"]

    out_fields = ["out_file"]

    # input
    intnorm_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                               name="intnorm_input")

    # fix the affine matrix (it's necessary for some cases)
    resample = setup_node(Function(function=resample_to_img,
                                   input_names=["in_file", "target", "interpolation"],
                                   output_names=["out_file"],
                                   imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                          name="resample_mask")
    resample.inputs.interpolation = "nearest"

    # calculate masked mean value
    mean_val = setup_node(Function(function=math_img,
                                   input_names=["formula", "img", "mask"],
                                   output_names=["out_value"],
                                   imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                          name='mean_value')
    mean_val.inputs.formula = "np.mean(np.nonzero(img[mask > 0]))"

    # normalize
    norm_img = setup_node(Function(function=math_img,
                                   input_names=["formula", "out_file", "img", "val"],
                                   output_names=["out_file"],
                                   imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                          name='norm_img')
    norm_img.inputs.formula = "img / val"

    # output
    intnorm_output = setup_node(IdentityInterface(fields=out_fields),
                                name="intnorm_output")

    # Create the workflow object
    wf = Workflow(name=wf_name)

    wf.connect([
                # resample
                (intnorm_input, resample,  [("source",    "target"),
                                            ("mask",      "in_file")]),

                # normalize
                (intnorm_input, mean_val,  [("source",    "img" )]),
                (resample,      mean_val,  [("out_file",  "mask")]),

                (intnorm_input, norm_img,  [("source",    "img"),
                                            (("source", rename, "_intnormed"), "out_file"),
                                           ]),

                (mean_val,      norm_img,  [("out_value", "val")]),
                (norm_img, intnorm_output, [("out_file",  "out_file")]),
               ])

    return wf




# -*- coding: utf-8 -*-
"""
Utilities to help in the DTI pre-processing
"""
from nipype.algorithms.rapidart import ArtifactDetect


def rapidart_dti_artifact_detection():
    art = ArtifactDetect()
    art.inputs.use_differences      = [True, False]
    art.inputs.use_norm             = True
    art.inputs.zintensity_threshold = 2
    art.inputs.norm_threshold       = 1
    art.inputs.mask_type            = 'file'
    art.inputs.parameter_source     = 'FSL'
    return art


def nlmeans_denoise(in_file, mask_file, out_file='', N=12):
    """ Filepath interface to the nlmeans_denoise_img in neuro_pypes.preproc."""
    import os.path as op
    import nibabel as nib

    from neuro_pypes.preproc import nlmeans_denoise_img
    from neuro_pypes.utils import rename

    den = nlmeans_denoise_img(nib.load(in_file), mask=nib.load(mask_file), N=N)

    if not out_file:
        out_file = rename(in_file, '_denoised')

    den.to_filename(out_file)

    return op.abspath(out_file)


def reslice(in_file, new_zooms=None, order=3, out_file=''):
    """
    Performs regridding of an image to set isotropic voxel sizes using dipy.
    """
    import os.path as op

    import nibabel as nib

    from neuro_pypes.preproc import reslice_img
    from neuro_pypes.utils import rename

    img = reslice_img(nib.load(in_file), new_zooms=new_zooms, order=order)

    if not out_file:
        out_file = rename(in_file, '_resliced')

    img.to_filename(out_file)

    return op.abspath(out_file)


def dti_acquisition_parameters(in_file, epi_factor=128):
    """ Return the path to the acqp_file and index file necessary for the new Eddy tool.
    This works when `in_file` is a nifti file obtained from a recent version from dcm2nii.

    This assumes a standard Siemens acquisition if the 'descrip.phaseDir' field is not
    in the nifti header.

    Please check the InPlanePhaseEncodingDirection field in the DICOM files, if its values is
    'COL', you are good to go.

    Notes
    -----
    # Comments on the `eddy` tool from FSL FDT.

    A description of the tool:
    http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/Eddy/UsersGuide

    Our problem to run this tool instead of the good-old `eddy_correct` is the `--acqp` argument, an
    acquisitions parameters file.

    A detailed description of the --acpq input file is here:
    http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddy/Faq#How_do_I_know_what_to_put_into_my_--acqp_file

    In the following subsections I describe each of the fields needed to check and build the acquisitions parameters
    file.

    ## Phase Encoding Direction
    Dicom header field: (0018,1312) InPlanePhaseEncodingDirection
    The phase encoding direction is the OPPOSITE of frequency encoding direction:
    - 'COL' = A/P (freqDir L/R),
    - 'ROW' = L/R (freqDir A/P)


    Nifti header field: "descrip.phaseDir:'+'" is for 'COL' in the DICOM InPlanePhaseEncodingDirection value.
    So if you have only one phase encoding oriendation and a '+' in the previous header field,
    the first 3 columns for the `--acqp` parameter file should be:
    0 1 0

    indicating that the scan was acquired with phase-encoding in the anterior-posterior direction.

    For more info:
    http://web.stanford.edu/group/vista/cgi-bin/wiki/index.php/DTI_Preprocessing_User_Manual#Frequency_vs_Phase_Encode_Direction
    https://xwiki.nbirn.org:8443/xwiki/bin/view/Function-BIRN/PhaseEncodeDirectionIssues


    ## Effective Echo Spacing (aka dwell time)

    Effective Echo Spacing (s) = 1/(BandwidthPerPixelPhaseEncode * MatrixSizePhase)

    effective echo spacing = 1 / [(0019,1028) * (0051,100b component #1)] (from the archives)
    https://www.jiscmail.ac.uk/cgi-bin/webadmin?A3=ind1303&L=FSL&E=quoted-printable&P=29358&B=--B_3444933351_15386849&T=text%2Fhtml;%20charset=ISO-8859-1&pending=

    The dwell time is in the nifti header `descrip.dwell`, in seconds (or look at the field `time_units`).
    http://www.mit.edu/~satra/nipype-nightly/interfaces/generated/nipype.interfaces.fsl.epi.html

    More info:
    http://lcni.uoregon.edu/kb-articles/kb-0003

    ## EPI factor

    The EPI factor is not included in the nifti header.
    You can read it using the Grassroots DICOM tool called `gdcmdump`, for example:
    >>> gdcmdump -C IM-0126-0001.dcm | grep 'EPIFactor'
    sFastImaging.lEPIFactor                  = 128

    More info:
    http://dicomlookup.com/default.htm


    # The fourth element of the acquisitions parameter file

    The fourth element in each row is the time (in seconds) between reading the center of the first echo and reading the
    center of the last echo.
    It is the "dwell time" multiplied by "number of PE steps - 1" and it is also the reciprocal of the PE bandwidth/pixel.

    Total readout time (FSL) = (number of echoes - 1) * echo spacing

    Total Readout Time (SPM) = 1/(BandwidthPerPixelPhaseEncode)
    Since the Bandwidth Per Pixel Phase Encode is in Hz, this will give the readout time in seconds


    # The `---index` argument

    The index argument is a text file with a row of numbers. Each number
    indicates what line (starting from 1) in the `acqp` file corresponds to
    each volume in the DTI acquisition.

    # What to do now with the `dcm2nii` files?

    I see two options to calculate the `acqp` lines with these files.

    1. We already have the `dwell` but we don't have the EPI factor.
    We know that the standard in Siemens is 128 and we could stick to that.

    2. Use the `slice_duration * 0.001` which is very near the calculated value.

    # Summary

    So, for example, if we had these acquisition parameters:

    ```
    Phase enc. dir. P >> A
    Echo spacing 0.75 [ms]
    EPI factor 128
    ```

    We should put in the `acqp` file this line:
    0 1 0 0.095
    """
    # the imports must be inside if you want them to work in a nipype.Function node.
    import warnings
    import os.path as op
    import nibabel as nib

    acqp_file = "diff.acqp"
    index_file = "diff.index"

    image = nib.load(in_file)
    n_volumes = image.shape[-1]

    # read the value of the descrip field in the nifti header
    descrip_field = image.header["descrip"].astype(str)[()]
    descrip = dict([item.split("=", 1) for item in descrip_field.split(";")])

    # assume by default phaseDir = '+'
    phase_dir = descrip.get("phaseDir", '+')
    pe_axis   = "0 1 0"
    if phase_dir == "-":
        pe_axis = "0 -1 0"

    if 'phaseDir' not in descrip:
        warnings.warn("Could not find or understand the value for phaseDir: {}. "
                      "Using default PE axis {}.".format(descrip.get("phaseDir"), pe_axis),
                      category=RuntimeWarning, stacklevel=2)

    # (number of phase-encode steps - 1) *
    # (echo spacing time in milliseconds) *
    # (seconds per millisecond)
    total_readout_time = (epi_factor - 1) * float(descrip["dwell"]) * 1e-3

    with open(acqp_file, "wt") as fout:
        fout.write("{} {}\n".format(pe_axis, total_readout_time))

    with open(index_file, "wt") as fout:
        fout.write("{}\n".format(" ".join(n_volumes * ["1"])))

    return op.abspath(acqp_file), op.abspath(index_file)
